import logging
import numpy as np
import pandas as pd

from abc import ABC, abstractmethod
from datetime import datetime

log = logging.getLogger(__name__)

group_reader_map = {}


class GroupReader(ABC):
    """
      ABC for all HDF Group Reader types.
      Each subclass must define a class attribute of <subclass_type_key>
      which contains a string identifying the Group type handled by that
      reader.
      These attributes then automatically populate the group_reader_map
      when a subclass is defined.  This map is then used by the adapter
      to instantiate the correct reader for the intercal format of a
      given group.
    """

    subclass_type_key = "pandas_type"

    def __init_subclass__(cls, /, **kwargs):
        super().__init_subclass__(**kwargs)
        keyattr = __class__.subclass_type_key
        reader_type = getattr(cls, keyattr, None)
        if not reader_type or not isinstance(reader_type, str):
            raise NotImplementedError(f"Reader subclass {cls.__name__} does not define a '{keyattr}' attribute")
        group_reader_map[reader_type] = cls
        log.debug(f"Registered HDF Group reader '{cls.__name__}' for type '{reader_type}'")

    @abstractmethod
    def get_group_shape(self):
        ...

    @abstractmethod
    def get_columns_of_group(self):
        ...

    @abstractmethod
    def get_index_info(self):
        ...

    @abstractmethod
    def get_series_info(self, series):
        ...

    @abstractmethod
    def get_index_range(self, start=0, end=None):
        ...

    @abstractmethod
    def get_columns_as_dataframe(self, columns, start=0, end=None):
        ...


class FrameGroupReader(GroupReader):
    """
    Reader for groups where the `pandas_type` attribute of the HDF group is "frame".
    This corresponds to a group where the application-level DataFrame is decomposed
    into `axis` and `block` components which are described by a collection of
    attributes and datasets inside the HDF group.
    The base class abstract methods are overridden in this class to provide
    implementations which deal specifically with groups of this format.
    """
    pandas_type = "frame"

    def __init__(self, hf, groupname):
        self.hf = hf
        try:
            self.group = hf[groupname]
        except KeyError as ke:
            raise ValueError(f"Error: file {hf.filename} contains no group {groupname}") from ke

    def get_columns_of_group(self):
        try:
            columns_raw = [*self.group["axis0"]]
        except KeyError as ke:
            raise ValueError(f"Error: {self.group.name} has invalid format") from ke
        return [col.decode() if isinstance(col, bytes) else col for col in columns_raw]

    def get_columns_by_block(self):
        try:
            nblocks = self.group.attrs["nblocks"]
        except KeyError as ke:
            raise KeyError(f"Error: group {self.group.name} contains no blocks") from ke
        blocks_columns = []
        for block_idx in range(nblocks):
            columns_raw = [*self.group[f"block{block_idx}_items"]]
            blocks_columns.append([col.decode() if isinstance(col, bytes) else col for col in columns_raw])
        return blocks_columns

    def make_column_map_of_group(self, blocks):
        group_cols = self.get_columns_of_group()
        return [column_to_block_coord(column, blocks) for column in group_cols]

    def get_series_by_column_names(self, column_names, start=None, end=None):
        columns = self.get_columns_of_group()
        block_columns = self.get_columns_by_block()
        if not column_names:
            # Return all columns when <column_names> arg
            # is either None or any empty container
            column_names = columns

        column_map = self.make_column_map_of_group(block_columns)
        named_map = {cname: cmap for cname, cmap in zip(columns, column_map)}

        row_sz, _ = self.get_group_shape()

        start = start or 0
        end = end or row_sz
        if start < 0 or start >= row_sz or start >= end or end < 0 or end >= row_sz+1:
            raise ValueError(f"Invalid bounds ({start=}, {end=} for series of length {row_sz}")

        column_series = {}
        for column_name in column_names:
            try:
                block_idx, col_idx = named_map[column_name]
            except KeyError as ke:
                raise ValueError(f"Group {self.group.name} contains no series '{column_name}'") from ke
            block_values_name = f"block{block_idx}_values"
            block_values = self.group[block_values_name]
            section = np.array([row[col_idx] for row in block_values[start:end]])
            column_series[column_name] = section

        return column_series

    def find_index_axis_index(self):
        # If index is a Pandas type, an "index_class" attr
        # will be present to indicate the index
        for ent in self.group:
            try:
                index_class = self.group[ent].attrs["index_class"]
                return ent
            except KeyError:
                continue

        # No item has a designated index, assume axis1 is index,
        # as per non-Datetime/Period Pandas index
        if "axis1" in self.group:
            return "axis1"

        # Otherwise, no suitable index can be found
        raise ValueError(f"Group {self.group.name} of pandas_type "
                          "'frame' contains no index axis")

    def get_index_range(self, start=None, end=None):
        index_axis = self.find_index_axis_index()
        index = self.group[index_axis]
        try:
            index_class = index.attrs["index_class"].decode()
        except KeyError:
            index_class = ""

        start = start or 0
        end = end or len(index)
        if start < 0 or start >= len(index) or start >= end or end < 0 or end >= len(index)+1:
            raise ValueError(f"Invalid bounds ({start=}, {end=} for index of length {len(index)}")

        if index_class.lower() == "datetime":
            return [str(nscale(row)) for row in index[start:end]]
        else:
            return index[start:end]

    def get_columns_as_dataframe(self, columns, start=None, end=None):
        index_range = self.get_index_range(start, end)
        index_range = [i.decode() if isinstance(i, bytes) else i for i in index_range]
        column_series = self.get_series_by_column_names(columns, start, end)
        return make_pandas_dataframe(index_range, column_series)

    def get_group_shape(self):
        index_axis = self.find_index_axis_index()
        row_sz = len(self.group[index_axis])
        col_sz = len(self.get_columns_of_group())
        return (row_sz, col_sz)

    def get_index_info(self):
        index_axis = self.find_index_axis_index()
        index = self.group[index_axis]
        try:
            dtype = index.attrs["kind"].decode()
        except KeyError:
            dtype = ""
        try:
            name = index.attrs["name"].decode()
        except KeyError:
            name = ""

        return {
            "name": name,
            "length": len(index),
            "dtype": dtype
        }

    def get_series_info(self, column_name):
        columns = self.get_columns_of_group()
        block_columns = self.get_columns_by_block()

        column_map = self.make_column_map_of_group(block_columns)
        named_map = {cname: cmap for cname, cmap in zip(columns, column_map)}

        block_idx, col_idx = named_map[column_name]
        block_values_name = f"block{block_idx}_values"
        block_values = self.group[block_values_name]

        return {
            "name": column_name,
            "length": block_values.shape[0],
            "dtype": str(block_values.dtype)
        }


class FrameTableGroupReader(GroupReader):
    """
    Reader for groups where the `pandas_type` attribute of the HDF group is "frame_table".
    This category of group stores the description of the layout of a DataFrame over
    attributes of the group itself and of a `table` dataset contained in the group.
    The "index_cols" attribute of a group stores the set of column names and the
    `FIELD_x` and `values_block_x` attributes of the dataset indicate in which of the
    dtype-specific columns of the dataset a particular series can be found.
    The base class abstract methods are overridden in this class to provide
    implementations which deal specifically with groups of this format.
    """
    pandas_type = "frame_table"

    def __init__(self, hf, groupname):
        self.hf = hf
        try:
            self.group = hf[groupname]
        except KeyError as ke:
            raise ValueError(f"Error: file {hf.filename} contains no group {groupname}") from ke
        try:
            self.table = self.group["table"]
        except KeyError as ke:
            raise ValueError(f"Error: file {hf.filename} has invalid format, no 'table'") from ke

    def get_index_column_index(self):
        max_index_depth = 128
        try:
            index_cols_raw = self.group.attrs["index_cols"].decode().split('\n')[2:-2:2]
        except KeyError:
            # No index columns present
            return None
        index_cols = trim_x_first_y_rest(1, 2, index_cols_raw)
        index_name = index_cols[0]  # Assume single index

        for field_idx in range(max_index_depth):
            try:
                field_name = self.table.attrs[f"FIELD_{field_idx}_NAME"]
            except KeyError:
                # No field has the name of the stated index column
                return None
            if field_name.decode() == index_name:
                return field_idx

    def get_values_start_block_index(self):
        max_value_depth = 128
        try:
            value_cols_raw = self.group.attrs["values_cols"].decode().split('\n')[1:-1:2]
        except KeyError:
            # Group contains no value columns
            return None
        value_cols = trim_x_first_y_rest(1, 2, value_cols_raw)
        first_value_col = value_cols[0]

        for field_idx in range(max_value_depth):
            try:
                field_name = self.table.attrs[f"FIELD_{field_idx}_NAME"]
            except KeyError:
                # No more fields
                return None
            if field_name.decode() == first_value_col:
                return field_idx

    def get_index_range(self, start=None, end=None):
        start = start or 0
        end = end or len(self.table)-1
        if start < 0 or start >= len(self.table) or start >= end or end < 0 or end >= len(self.table):
            raise ValueError(f"Invalid bounds ({start=}, {end=} for series of length {len(self.table)}")

        index_idx = self.get_index_column_index()
        try:
            index_name = self.table.attrs[f"FIELD_{index_idx}_NAME"].decode()
            index_kind = self.table.attrs[f"{index_name}_kind"].decode()
        except KeyError:
            return []

        if index_idx is None:
            return []
        if index_kind.lower().find("datetime") != -1:
            return [str(nscale(row[index_idx])) for row in self.table[start:end]]
        else:
            return [row[index_idx] for row in self.table[start:send]]

    def get_columns_by_block(self):
        value_cols_raw = self.group.attrs["values_cols"].decode().split('\n')[1:-1:2]
        value_cols = trim_x_first_y_rest(1, 2, value_cols_raw)

        block_columns = []
        for field_idx, field_name in enumerate(value_cols, start=1):
            kind = self.table.attrs[f"{field_name}_kind"]
            kind_cols_raw = kind.decode().split('\n')[1:-1:2]
            kind_cols = trim_x_first_y_rest(1, 2, kind_cols_raw)
            block_columns.append(kind_cols)

        return block_columns

    def get_series_by_column_names(self, column_names, start=None, end=None):
        columns = self.get_columns_of_group()
        block_columns = self.get_columns_by_block()

        column_map = self.make_column_map_of_group(block_columns)
        named_map = {cname: cmap for cname, cmap in zip(columns, column_map)}

        if not column_names:
            # Return all columns when <column_names> arg
            # is either None or any empty container
            column_names = columns

        first_value_block_idx = self.get_values_start_block_index()

        start = start or 0
        end = end or len(self.table)
        if start < 0 or start >= len(self.table) or start >= end or end < 0 or end >= len(self.table)+1:
            raise ValueError(f"Invalid bounds ({start=}, {end=} for series of length {len(self.table)}")

        column_series = {}
        for column_name in column_names:
            block_idx, col_idx = named_map[column_name]
            block_rows = np.array([row[block_idx+first_value_block_idx] for row in self.table[start:end]])
            section = block_rows[:, col_idx]
            column_series[column_name] = section

        return column_series

    def get_columns_as_dataframe(self, columns, start=None, end=None):
        index_range = self.get_index_range(start, end)
        column_series = self.get_series_by_column_names(columns, start, end)
        return make_pandas_dataframe(index_range, column_series)

    def make_column_map_of_group(self, blocks):
        non_index_cols = self.get_columns_of_group()
        return [column_to_block_coord(column, blocks) for column in non_index_cols]

    def get_columns_of_group(self):
        non_index_cols_raw = self.group.attrs["non_index_axes"].decode().split("\n")[3:-3:2]
        return trim_x_first_y_rest(1, 2, non_index_cols_raw)

    def get_group_shape(self):
        row_sz = len(self.table)
        col_sz = len(self.get_columns_of_group())
        return (row_sz, col_sz)

    def get_index_info(self):
        index_idx = self.get_index_column_index()
        index_name = self.table.attrs[f"FIELD_{index_idx}_NAME"].decode()
        index_kind = self.table.attrs[f"{index_name}_kind"].decode()
        index_rows = self.table.attrs["NROWS"]

        return {
            "name": index_name,
            "length": index_rows,
            "dtype": index_kind
        }

    def get_series_info(self, column_name):
        columns = self.get_columns_of_group()
        block_columns = self.get_columns_by_block()

        column_map = self.make_column_map_of_group(block_columns)
        named_map = {cname: cmap for cname, cmap in zip(columns, column_map)}

        block_idx, col_idx = named_map[column_name]
        dtype = self.table.attrs[f"values_block_{block_idx}_dtype"]

        return {
            "name": column_name,
            "length": len(self.table),
            "dtype": dtype.decode()
        }


"""
  Auxiliary Functions
"""

def nscale(ts):
    """
      Transforms integers representing nanoseconds past the epoch
      into instances of datetime.timestamp
    """
    return datetime.fromtimestamp(ts/1e9)

def trim_x_first_y_rest(x, y, names):
    return [names[0][x:], *[name[y:] for name in names[1:]] ]

def column_to_block_coord(column, blocks):
    for block_idx, block in enumerate(blocks):
        try:
            col_idx = block.index(column)
        except ValueError:
            continue
        return (block_idx, col_idx)
    raise ValueError(f"Column {column} not found in any block")

def make_pandas_dataframe(index, series):
    return pd.DataFrame(
        {name: values for name, values in series.items()},
        index=pd.Index(index)
    )



"""template

Revision ID: a7896842f484
Revises: f579b06e9cc2
Create Date: 2018-04-13 08:44:21.634411

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'a7896842f484'
down_revision = 'f579b06e9cc2'
branch_labels = None
depends_on = None

import logging
log = logging.getLogger(__name__)


def upgrade():
    if op.get_bind().dialect.name == 'mysql':

        # ## tTemplate

        try:
            op.alter_column('tTemplate', 'template_id', new_column_name='id', existing_type=sa.Integer(), autoincrement=True, nullable=False)
            op.alter_column('tTemplate', 'template_name', new_column_name='name', existing_type=sa.String(60), nullable=False)
        except Exception as e:
            log.exception(e)

        # ## tTemplateType

        try:
            op.alter_column('tTemplateType', 'type_id', new_column_name='id', existing_type=sa.Integer(), autoincrement=True, nullable=False)
            op.alter_column('tTemplateType', 'type_name', new_column_name='name', existing_type=sa.String(60), nullable=False)
        except Exception as e:
            log.exception(e)

    else: ## sqlite

        # ## tTemplate
        try:
            op.drop_table('tTemplate_new')
        except:
            log.info("tTemplate_new isn't there")

        try:
            # ## tTemplate
            op.create_table(
                'tTemplate_new',
                sa.Column('id', sa.Integer, primary_key=True),
                sa.Column('name', sa.Text(60), nullable=False, unique=True),
                sa.Column('cr_date', sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
                sa.Column('layout', sa.Text(1000)),
            )

            op.execute("insert into tTemplate_new (id, name, cr_date, layout) select template_id, template_name, cr_date, layout from tTemplate")

            op.rename_table('tTemplate','tTemplate_old')
            op.rename_table('tTemplate_new', 'tTemplate')
            op.drop_table('tTemplate_old')

        except Exception as e:
            log.exception(e)

        # ## tTemplateType
        try:
            op.drop_table('tTemplateType_new')
        except:
            log.info("tTemplateType_new isn't there")

        try:
            # ## tTemplateType
            op.create_table(
                'tTemplateType_new',
                sa.Column('id', sa.Integer(), primary_key=True, nullable=False),
                sa.Column('name', sa.String(60),  nullable=False, unique=True),
                sa.Column('template_id', sa.Integer(), sa.ForeignKey('tTemplate.id'), nullable=False),
                sa.Column('resource_type', sa.String(60)),
                sa.Column('alias', sa.String(100)),
                sa.Column('layout', sa.Text(1000)),
                sa.Column('cr_date', sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
                sa.UniqueConstraint('template_id', 'name', 'resource_type', name="unique type name"),
            )
            op.execute("insert into tTemplateType_new (name, id, template_id, resource_type, alias, layout, cr_date) select type_name, type_id, template_id, resource_type, alias, layout, cr_date from tTemplateType")

            op.rename_table('tTemplateType','tTemplateType_old')
            op.rename_table('tTemplateType_new', 'tTemplateType')
            op.drop_table('tTemplateType_old')

        except Exception as e:
            log.exception(e)

def downgrade():
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

from ..import config

import os
import subprocess 
import sys

import logging
log = logging.getLogger(__name__)

from lxml import etree

def get_plugins(**kwargs):
    """
        Get all available plugins
    """
    
    plugins = []
    plugin_paths = []
    
    #Look in directory or set of directories for
    #plugins
    
    base_plugin_dir = config.get('plugin', 'default_directory')
    plugin_xsd_path = config.get('plugin', 'plugin_xsd_path')

    base_plugin_dir_contents = os.listdir(base_plugin_dir)
    for directory in base_plugin_dir_contents:
        #ignore hidden files
        if directory[0] == '.' or directory == 'xml':
            continue

        #Is this a file or a directory? If it's a directory, it's a plugin.
        path = os.path.join(base_plugin_dir, directory)
        if os.path.isdir(path):
            plugin_paths.append(path)
   

    #For each plugin, get its details (an XML string)
    
    #Retrieve the xml schema for validating the XML to make sure
    #what is being provided  to the IU is correct.
    xmlschema_doc = etree.parse(plugin_xsd_path)
                    
    xmlschema = etree.XMLSchema(xmlschema_doc)
    
    #Get the xml description file from the plugin directory. If there
    #is no xml file, the plugin in unusable.
    for plugin_dir in plugin_paths:
        full_plugin_path = os.path.join(plugin_dir, 'trunk')
        
        dir_contents = os.listdir(full_plugin_path)
        #look for a plugin.xml file in the plugin directory 
        for file_name in dir_contents:
            file_path = os.path.join(full_plugin_path, file_name)
            if file_name == 'plugin.xml':
                
                f = open(file_path, 'r')
                
                #validate the xml using the xml schema for defining
                #plugin details
                try:
                    y = open(file_path, 'r')
                    
                    xml_tree = etree.parse(y)

                    xmlschema.assertValid(xml_tree)
                   
                    plugins.append(etree.tostring(xml_tree))
                except Exception as e:
                    log.critical("Schema %s did not validate! (error was %s)"%(file_name, e))

                break
        else:
            log.warning("No xml plugin details found for %s. Ignoring", plugin_dir)

    return plugins

def run_plugin(plugin,**kwargs):
    """
        Run a plugin
    """
   
    args = [sys.executable]

    #Get plugin executable
    home = os.path.expanduser('~')
    path_to_plugin = os.path.join(home, 'svn/HYDRA/HydraPlugins', plugin.location)
    args.append(path_to_plugin)
    
    #Parse plugin arguments into a string
    plugin_params = " "
    for p in plugin.params:
        param = "--%s=%s "%(p.name, p.value)
        args.append("--%s"%p.name)
        args.append(p.value)
        plugin_params = plugin_params + param

    log_dir = config.get('plugin', 'result_file')
    log_file = os.path.join(home, log_dir, plugin.name)

    #this reads all the logs so far. We're not interested in them
    #Everything after this is new content to the file
    try:
        f = open(log_file, 'r')
        f.read()
    except:
        f = open(log_file, 'w')
        f.close()
        f = open(log_file, 'r')

    pid = subprocess.Popen(args).pid
    #run plugin
    #os.system("%s %s"%(path_to_plugin, plugin_params))
    
    log.info("Process started! PID: %s", pid)

    return str(pid)

   
    #monitor plugin

    #when plugin is done, read its output from 
    #the output file and return to the client


def check_plugin_status(plugin_name, pid,**kwargs):
    home = os.path.expanduser('~')
    log_dir = config.get('plugin', 'result_file')
    log_file = os.path.join(home, log_dir, plugin_name)
    try:
        f = open(log_file, 'r')
        file_text = f.read()
        pid_index = file_text.find("%%%s%%"%(pid))

        if pid_index < 0:
            return "No log found for PID %s in %s"%(pid, plugin_name)

        split_string = file_text.split("%%%s%%"%(pid))

        return split_string[1]

    except IOError as e:
        return "No log file found for %s in plugin %s Error was: %s"%(pid, plugin_name, e)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013, 2014, University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#
import os
import glob
import sys

PYTHONVERSION = sys.version_info
if PYTHONVERSION >= (3,2):
    import configparser as ConfigParser
else:
    import ConfigParser
import logging

global CONFIG
CONFIG = None

global localfiles
global localfile
global repofile
global repofiles
global userfile
global userfiles
global sysfile
global sysfiles

def load_config():
    """Load a config file. This function looks for a config (*.ini) file in the
    following order::

        (1) ./*.ini
        (2) ~/.config/hydra/
        (3) /etc/hydra
        (4) /path/to/hydra_base/*.ini

    (1) will override (2) will override (3) will override (4). Parameters not
    defined in (1) will be taken from (2). Parameters not defined in (2) will
    be taken from (3).  (3) is the config folder that will be checked out from
    the svn repository.  (2) Will be be provided as soon as an installable
    distribution is available. (1) will usually be written individually by
    every user."""
    global localfiles
    global localfile
    global repofile
    global repofiles
    global userfile
    global userfiles
    global sysfile
    global sysfiles
    global CONFIG
    logging.basicConfig(level='INFO')

    config = ConfigParser.ConfigParser(allow_no_value=True)

    modulepath = os.path.dirname(os.path.abspath(__file__))

    localfile = os.path.join(os.getcwd(), 'hydra.ini')
    localfiles = glob.glob(localfile)

    repofile = os.path.join(modulepath, 'hydra.ini')
    repofiles = glob.glob(repofile)

    if sys.platform.startswith("win"):
        from hydra_base.util.windows import win_get_common_documents
        userfile = os.path.join(os.path.expanduser('~'),'AppData','Local','hydra.ini')
        userfiles = glob.glob(userfile)

        sysfile = os.path.join(win_get_common_documents(), 'Hydra','hydra.ini')
        sysfiles = glob.glob(sysfile)
    else:
        userfile = os.path.join(os.path.expanduser('~'), '.hydra', 'hydra.ini')
        userfiles = glob.glob(userfile)

        sysfile = os.path.join('etc','hydra','hydra.ini')
        sysfiles = glob.glob(sysfile)


    for ini_file in repofiles:
        logging.debug("Repofile: %s"%ini_file)
        config.read(ini_file)
    for ini_file in sysfiles:
        logging.debug("Sysfile: %s"%ini_file)
        config.read(ini_file)
    for ini_file in userfiles:
        logging.debug("Userfile: %s"%ini_file)
        config.read(ini_file)
    for ini_file in localfiles:
        logging.info("Localfile: %s"%ini_file)
        config.read(ini_file)

    env_value = os.environ.get('HYDRA_CONFIG')
    if env_value is not None:
        if os.path.exists(env_value):
            config.read(env_value)
        else:
            logging.warning('HYDRA_CONFIG set as %s but file does not exist', env_value)


    try:
        home_dir = config.get('DEFAULT', 'home_dir')
    except:
        home_dir = os.environ.get('HYDRA_HOME_DIR', '~')
    config.set('DEFAULT', 'home_dir', os.path.expanduser(home_dir))

    try:
        hydra_base = config.get('DEFAULT', 'hydra_base_dir')
    except:
        hydra_base = os.environ.get('HYDRA_BASE_DIR', modulepath)
    config.set('DEFAULT', 'hydra_base_dir', os.path.expanduser(hydra_base))

    read_values_from_environment(config, 'mysqld', 'server_name')


    CONFIG = config

    return config

def read_values_from_environment(config, section_key, options_key):
    #####################################
    # Settings for docker ENV variables #
    #####################################
    env_var_name='HYDRA_DOCKER__' + section_key + '__' + options_key

    env_value = os.environ.get(env_var_name, '-')
    if (env_value != '-'):
        # Substitute the server_name with the end variable
        # print("Presente")
        config.set(section_key, options_key, env_value)


def get(section, option, default=None):

    if CONFIG is None:
        load_config()

    try:
        return CONFIG.get(section, option)
    except:
        return default

def getint(section, option, default=None):

    if CONFIG is None:
        load_config()

    try:
        return CONFIG.getint(section, option)
    except:
        return default



"""project_appdata

Revision ID: 877adf863b33
Revises: 04e4ae80b7b9
Create Date: 2023-08-14 16:03:11.415679

"""
import logging
from alembic import op
import sqlalchemy as sa

log = logging.getLogger(__name__)

# revision identifiers, used by Alembic.
revision = '877adf863b33'
down_revision = '04e4ae80b7b9'
branch_labels = None
depends_on = None

def upgrade():
    if op.get_bind().dialect.name == 'mysql':

        try:
            op.add_column('tProject',
                          sa.Column('appdata', sa.JSON()))
        except Exception as e:
            log.critical(e)


def downgrade():
    if op.get_bind().dialect.name == 'mysql':
        try:
            op.drop_column('tProject', 'appdata')
        except Exception as e:
            log.critical(e)


def upgrade():
    pass


def downgrade():
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

from functools import wraps
from .. import db
from ..db.model import Perm, User, Role, RolePerm, RoleUser
from sqlalchemy.orm.exc import NoResultFound
from ..exceptions import PermissionError



def check_perm(user_id, permission_code):
    """
        Checks whether a user has permission to perform an action.
        The permission_code parameter should be a permission contained in tPerm.

        If the user does not have permission to perfom an action, a permission
        error is thrown.
    """
    try:
        perm = db.DBSession.query(Perm).filter(Perm.code==permission_code).one()
    except NoResultFound:
        raise PermissionError("Nonexistent permission type: %s"%(permission_code))


    try:
        #get all the roles where the specified user has the specified permission
        qry = db.DBSession.query(RoleUser.role_id)\
            .join(RolePerm, RolePerm.role_id == RoleUser.role_id)\
            .filter(RolePerm.perm_id == perm.id)\
            .filter(RoleUser.user_id == user_id)

        res = qry.all()
    except NoResultFound:
        raise PermissionError("Permission denied. User %s does not have permission %s"%
                        (user_id, permission_code))

def check_role(user_id, role_code):
    """
        Checks whether a user has been assigned the specified role
        The role_code parameter should be a role code contained in tRole.

        If the user does not have code a permission error is thrown.
    """
    try:
        role = db.DBSession.query(Role).filter(Role.code==role_code).one()
    except NoResultFound:
        raise PermissionError("Nonexistent role type: %s"%(role_code))


    try:
        #get all the roles where the specified user has the specified permission
        qry = db.DBSession.query(RoleUser.role_id)\
            .filter(RoleUser.role_id == role.id)\
            .filter(RoleUser.user_id == user_id)

        res = qry.all()
    except NoResultFound:
        raise PermissionError("Permission denied. User %s does not have role %s"%
                        (user_id, role_code))



def required_perms(*req_perms):
    """
       Decorator applied to functions requiring caller to possess permission
       Takes args tuple of required perms and raises PermissionsError
       via check_perm if these are not a subset of user perms
    """
    def dec_wrapper(wfunc):
        @wraps(wfunc)
        def wrapped(*args, **kwargs):
            user_id = kwargs.get("user_id")

            #bind this here so that the 'updated by' columns can be updated
            #automatically in the DB, in order to ensure correct auditing
            db.DBSession.user_id = user_id

            for perm in req_perms:
                check_perm(user_id, perm)

            return wfunc(*args, **kwargs)

        return wrapped
    return dec_wrapper

def required_role(req_role):
    """
       Decorator applied to functions requiring caller to possess the specified role
    """
    def dec_wrapper(wfunc):
        @wraps(wfunc)
        def wrapped(*args, **kwargs):
            user_id = kwargs.get("user_id")
            try:
                res = db.DBSession.query(RoleUser).filter(RoleUser.user_id==user_id).join(Role).filter(Role.code==req_role).one()
            except NoResultFound:
                raise PermissionError("Permission denied. User %s does not have role %s"%
                                (user_id, req_role))

            return wfunc(*args, **kwargs)

        return wrapped
    return dec_wrapper



# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2020 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

import json
import logging
import re

from collections.abc import Sized

from hydra_base import db
from hydra_base.db.model import Attr
from hydra_base.lib.objects import JSONObject
from hydra_base.exceptions import HydraError
from hydra_base.util import dataset_util
from hydra_base.lib import units

log = logging.getLogger(__name__)

def get_attr(attr_id):
    attr = db.DBSession.query(Attr).filter(Attr.id == attr_id).one()
    return JSONObject(attr)

def check_dimension(typeattr, unit_id=None):
    """
        Check that the unit and dimension on a type attribute match.
        Alternatively, pass in a unit manually to check against the dimension
        of the type attribute
    """
    if unit_id is None:
        unit_id = typeattr.unit_id

    dimension_id = get_attr(typeattr.attr_id).dimension_id

    if unit_id is not None and dimension_id is None:
        # First error case
        unit_dimension_id = units.get_dimension_by_unit_id(unit_id).id
        unit = units.get_unit(unit_id)
        dimension = units.get_dimension(unit_dimension_id, do_accept_dimension_id_none=True)
        raise HydraError(f"Unit {unit_id} ({unit.abbreviation}) has dimension_id"+
                         f" {dimension.id}(name=dimension.name),"+
                         " but attribute has no dimension")
    elif unit_id is not None and dimension_id is not None:
        unit_dimension_id = units.get_dimension_by_unit_id(unit_id).id
        unit = units.get_unit(unit_id)
        dimension1 = units.get_dimension(unit_dimension_id, do_accept_dimension_id_none=True)
        dimension2 = units.get_dimension(unit_dimension_id, do_accept_dimension_id_none=True)
        if unit_dimension_id != dimension_id:
            # Only error case
            raise HydraError(f"Unit {unit_id} ({unit.abbreviation}) has dimension_id"+
                             f" {dimension1.id}(name=dimension1.name),"+
                             f" but attribute has id: {dimension2.id}({dimension2.name})")

def get_attr_by_name_and_dimension(name, dimension_id):
    """
        Search for an attribute with the given name and dimension_id.
        If such an attribute does not exist, create one.
    """

    attr = db.DBSession.query(Attr).filter(
        Attr.name == name,
        Attr.dimension_id == dimension_id).first()

    if attr is None:
        # In this case the attr does not exists so we must create it
        attr = Attr()
        attr.dimension_id = dimension_id
        attr.name = name

        log.debug("Attribute not found, creating new attribute: name:%s, dimen:%s",
                  attr.name, attr.dimension_id)

        db.DBSession.add(attr)

    return attr

def parse_data_restriction(restriction_dict):
    if restriction_dict is None or len(restriction_dict) == 0:
        return None

    #replace soap text with an empty string
    #'{soap_server.hydra_complexmodels}' -> ''
    dict_str = re.sub('{[a-zA-Z._]*}', '', str(restriction_dict))

    if isinstance(restriction_dict, dict):
        new_dict = restriction_dict
    else:
        new_dict = json.loads(restriction_dict)

    #Evaluate whether the dict actually contains anything.
    if not isinstance(new_dict, dict) or len(new_dict) == 0:
        log.critical('A restriction was specified, but it is null')
        return None

    ret_dict = {}
    for k, v in new_dict.items():
        if isinstance(v, Sized) and len(v) == 1:
            ret_dict[k] = v[0]
        else:
            ret_dict[k] = v

    return json.dumps(ret_dict)

def validate_resourcescenario(resourcescenario, template_id=None, **kwargs):
    """
        Perform a check to ensure a resource scenario's datasets are correct given what the
        definition of that resource (its type) specifies.
    """
    res = resourcescenario.resourceattr.get_resource()

    types = res.types

    dataset = resourcescenario.dataset

    if len(types) == 0:
        return

    if template_id is not None:
        if template_id not in [r.templatetype.template_id for r in res.types]:
            raise HydraError("Template %s is not used for resource attribute %s in scenario %s"%\
                             (template_id, resourcescenario.resourceattr.attr.name,
                             resourcescenario.scenario.name))

    #Validate against all the types for the resource
    for resourcetype in types:
        #If a specific type has been specified, then only validate
        #against that type and ignore all the others
        if template_id is not None:
            if resourcetype.templatetype.template_id != template_id:
                continue
        #Identify the template types for the template
        tmpltype = resourcetype.templatetype
        for ta in tmpltype.typeattrs:
            #If we find a template type which mactches the current attribute.
            #we can do some validation.
            if ta.attr_id == resourcescenario.resourceattr.attr_id:
                if ta.data_restriction:
                    log.debug("Validating against %s", ta.data_restriction)
                    validation_dict = json.loads(ta.data_restriction)
                    dataset_util.validate_value(validation_dict, dataset.get_val())

def validate_resource(resource, tmpl_types, resource_scenarios=[], **kwargs):
    errors = []
    resource_type = None

    #No validation required if the resource has no type.
    if len(resource.types) == 0:
        return []

    for rt in resource.types:
        if tmpl_types.get(rt.type_id) is not None:
            resource_type = tmpl_types[rt.type_id]
            break
        else:
            errors.append("Type %s not found on %s %s"%
                          (tmpl_types, resource_type, resource.get_name()))

    ta_dict = {}
    for ta in resource_type.typeattrs:
        ta_dict[ta.attr_id] = ta

    #Make sure the resource has all the attributes specified in the template
    #by checking whether the template attributes are a subset of the resource
    #attributes.
    type_attrs = set([ta.attr_id for ta in resource_type.typeattrs])

    resource_attrs = set([ra.attr_id for ra in resource.attributes])

    if not type_attrs.issubset(resource_attrs):
        for ta in type_attrs.difference(resource_attrs):
            errors.append("Resource %s does not have attribute %s"%
                          (resource.get_name(), ta_dict[ta].attr.name))

    resource_attr_ids = set([ra.id for ra in resource.attributes])
    #if data is included, check to make sure each dataset conforms
    #to the boundaries specified in the template: i.e. that it has
    #the correct dimension and (if specified) unit.
    if len(resource_scenarios) > 0:
        for ra_id in resource_attr_ids:
            rs = resource_scenarios.get(ra_id)
            if rs is None:
                continue
            attr_name = rs.resourceattr.attr.name
            rs_unit_id = rs.dataset.unit_id

            rs_dimension_id = units.get_dimension_by_unit_id(rs_unit_id, do_accept_unit_id_none=True).id

            type_dimension_id = ta_dict[rs.resourceattr.attr_id].attr.dimension_id
            type_unit_id = ta_dict[rs.resourceattr.attr_id].unit_id

            if rs_dimension_id != type_dimension_id:
                errors.append("Dimension mismatch on %s %s, attribute %s: "
                              "%s on attribute, %s on type"%
                             (resource.ref_key, resource.get_name(), attr_name,
                              rs_dimension_id, type_dimension_id))

            if type_unit_id is not None:
                if rs_unit_id != type_unit_id:
                    errors.append("Unit mismatch on attribute %s. "
                                  "%s on attribute, %s on type"%
                                 (attr_name, rs_unit_id, type_unit_id))
    if len(errors) > 0:
        log.warning(errors)

    return errors



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

import os
import json
from ..db.model import Network, Scenario, Project, User, Role, Perm, RolePerm, RoleUser, ResourceAttr, ResourceType, Dimension, Unit
from sqlalchemy.orm.exc import NoResultFound, MultipleResultsFound
from .. import db
import datetime
import random
import bcrypt
from ..exceptions import HydraError, HydraLoginUserNotFound, HydraLoginUserMaxAttemptsExceeded, HydraLoginUserPasswordWrong
from sqlalchemy.orm import load_only
from sqlalchemy import func
from ..lib.objects import JSONObject
from ..lib.users import get_remaining_login_attempts, inc_failed_login_attempts

import logging
log = logging.getLogger(__name__)


def add_resource_types(resource_i, types):
    """
    Save a reference to the types used for this resource.

    @returns a list of type_ids representing the type ids
    on the resource.

    """
    if types is None:
        return []

    existing_type_ids = []
    if resource_i.types:
        for t in resource_i.types:
            existing_type_ids.append(t.type_id)

    new_type_ids = []
    for templatetype in types:

        if templatetype.id in existing_type_ids:
            continue

        rt_i = ResourceType()
        rt_i.type_id     = templatetype.id
        rt_i.ref_key     = resource_i.ref_key
        rt_i.template_id = resource_i.child_template_id
        if resource_i.ref_key == 'NODE':
            rt_i.node_id      = resource_i.id
        elif resource_i.ref_key == 'LINK':
            rt_i.link_id      = resource_i.id
        elif resource_i.ref_key == 'GROUP':
            rt_i.group_id     = resource_i.id
        resource_i.types.append(rt_i)
        new_type_ids.append(templatetype.id)

    return new_type_ids

def add_resource_attributes(resource_i, attributes):
    if attributes is None:
        return {}
    resource_attrs = {}
    #ra is for ResourceAttr
    for ra in attributes:

        if ra.id < 0:
            ra_i = resource_i.add_attribute(ra.attr_id, ra.attr_is_var)
            db.DBSession.add(ra_i)
        else:
            ra_i = db.DBSession.query(ResourceAttr).filter(ResourceAttr.id==ra.id).one()
            ra_i.attr_is_var = ra.attr_is_var

        resource_attrs[ra.id] = ra_i
    return resource_attrs


def make_root_user():
    try:
        user = db.DBSession.query(User).filter(User.username=='root').one()
    except NoResultFound:

        root_pwd = bcrypt.hashpw(b'', bcrypt.gensalt())

        user = User(username='root',
                    password=root_pwd,
                    display_name='Root User')
        db.DBSession.add(user)

    try:
        role = db.DBSession.query(Role).filter(Role.code=='admin').one()
    except NoResultFound:
        raise HydraError("Admin role not found.")
    except MultipleResultsFound:
        roles = list(db.DBSession.query(Role).all())
        log.info(roles)
        role = roles[0]
        #raise HydraError("Multiple rows found")

    try:
        userrole = db.DBSession.query(RoleUser).filter(RoleUser.role_id==role.id,
                                                       RoleUser.user_id==user.id).one()
    except NoResultFound:
        userrole = RoleUser(role_id=role.id,user_id=user.id)
        user.roleusers.append(userrole)
        db.DBSession.add(userrole)


    db.DBSession.flush()

    user_id = user.id

    # Do not remove!
    db.commit_transaction()

    return user_id


def login_user(username, password):

    try:
        user_i = db.DBSession.query(User).filter(User.username == username).one()
    except NoResultFound:
        """ The user has not been found in the DB """
        raise HydraLoginUserNotFound(username)
    except:
        """ Generic DB Error """
        raise HydraError(username)

    if get_remaining_login_attempts(username, user_id=user_i.id) <= 0:
        """  Account is not permitted to login """
        raise HydraLoginUserMaxAttemptsExceeded("Max login attempts exceeded for user {}".format(username))

    userPassword = ""
    try:
        userPassword = user_i.password.encode('utf-8')
    except (AttributeError, UnicodeEncodeError):
        userPassword = user_i.password

    try:
        password = password.encode('utf-8')
    except (AttributeError, UnicodeEncodeError):
        pass

    if bcrypt.checkpw(password, userPassword):
        user_i.last_login = datetime.datetime.now()
        user_i.failed_logins = 0
        user_id = user_i.id
        db.DBSession.flush()
        # Do not commit the transaction here because it is managed by HWI/Hydra-Server
        # db.commit_transaction()
        return user_id
    else:
        log.info("User {} now has {} failed logins".format(username, user_i.failed_logins+1))
        inc_failed_login_attempts(user_i.username, user_id=1)
        raise HydraLoginUserPasswordWrong(username)


def create_default_net():
    try:
        net = db.DBSession.query(Network).filter(Network.id==1).one()
    except NoResultFound:
        project = Project(name="Project network", created_by=1)
        net = Network(name="Default network", created_by=1)
        scen = Scenario(name="Default network", created_by=1)
        project.networks.append(net)
        net.scenarios.append(scen)
        db.DBSession.add(net)
    db.DBSession.flush()
    # Do not remove!
    db.commit_transaction()
    return net


def create_default_users_and_perms():
    """
        Adds the roles and perm to the DB. It adds only roles, perms and links between them that are not inside the db
        It is possible adding new role or perm and connecting them just modifiying the following lists
    """

    # perms = db.DBSession.query(Perm).all()
    # if len(perms) > 0:
    #     return

    log.info("Adding default users and permissions")

    default_perms = (
        ("get_user", "Get User"),
        ("add_user", "Add User"),
        ("edit_user", "Edit User"),
        ("get_role", "Get Role"),
        ("add_role", "Add Role"),
        ("edit_role", "Edit Role"),
        ("get_perm", "Get Permission"),
        ("add_perm", "Add Permission"),
        ("edit_perm", "Edit Permission"),

        ("add_attribute", "Add Attribute"),
        ("get_attribute", "Get Attribute"),
        ("edit_attribute", "Edit Attribute"),
        ("delete_attribute", "Delete Attribute"),

        ("add_network", "Add network"),
        ("edit_network", "Edit network"),
        ("get_network", "Get network"),
        ("delete_network", "Delete network"),
        ("share_network", "Share network"),
        ("edit_topology", "Edit network topology"),

        ("get_project", "Get Project"),
        ("add_project", "Add Project"),
        ("edit_project", "Edit Project"),
        ("delete_project", "Delete Project"),
        ("share_project", "Share Project"),

        ("edit_data", "Edit network data"),
        ("get_data", "View network data"),

        ("add_template", "Add Template"),
        ("edit_template", "Edit Template"),
        ("get_template", "Get Template"),
        ("delete_template", "Delete Template"),

        ("add_dimension", "Add Dimension"),
        ("update_dimension", "Update Dimension"),
        ("delete_dimension", "Delete Dimension"),

        ("add_unit", "Add Unit"),
        ("update_unit", "Update Unit"),
        ("delete_unit", "Delete Unit"),

        ('get_rules', "View Rules"),
        ('add_rules', "Add Rules"),
        ('update_rules', "Edit Rules"),
        ('share_rules', "Share Rules"),
        ('delete_rules', "Delete Rules")

    )

    default_roles = (
        ("admin", "Administrator"),
        ("dev", "Developer"),
        ("modeller", "Modeller / Analyst"),
        ("manager", "Manager"),
        ("grad", "Graduate"),
        ("developer", "Developer"),
        ("decision", "Decision Maker"),
    )

    roleperms = (
        # Admin permissions
        ('admin', "get_user"),
        ('admin', "add_user"),
        ('admin', "edit_user"),
        ('admin', "get_role"),
        ('admin', "add_role"),
        ('admin', "edit_role"),
        ('admin', "get_perm"),
        ('admin', "add_perm"),
        ('admin', "edit_perm"),
        ('admin', "add_attribute"),
        ('admin', "edit_attribute"),
        ('admin', "get_attribute"),
        ('admin', "delete_attribute"),
        ('admin', "add_network"),
        ('admin', "edit_network"),
        ('admin', "get_network"),
        ('admin', "delete_network"),
        ('admin', "share_network"),
        ('admin', "get_project"),
        ('admin', "add_project"),
        ('admin', "edit_project"),
        ('admin', "delete_project"),
        ('admin', "share_project"),
        ('admin', "edit_topology"),
        ('admin', "edit_data"),
        ('admin', "get_data"),
        ('admin', "add_template"),
        ('admin', "edit_template"),
        ('admin', "get_template"),
        ('admin', "delete_template"),

        ('admin', "add_dimension"),
        ('admin', "update_dimension"),
        ('admin', "delete_dimension"),

        ('admin', "add_unit"),
        ('admin', "update_unit"),
        ('admin', "delete_unit"),

        ('admin', 'get_rules'),
        ('admin', 'add_rules'),
        ('admin', 'update_rules'),
        ('admin', 'share_rules'),
        ('admin', 'delete_rules'),

        # Developer permissions
        ('developer', "add_attribute"),
        ('developer', "edit_attribute"),
        ('developer', "get_attribute"),
        ('developer', "delete_attribute"),
        ("developer", "add_network"),
        ("developer", "edit_network"),
        ("developer", "delete_network"),
        ("developer", "share_network"),
        ('developer', "get_project"),
        ("developer", "add_project"),
        ("developer", "edit_project"),
        ("developer", "delete_project"),
        ("developer", "share_project"),
        ("developer", "edit_topology"),
        ("developer", "edit_data"),
        ("developer", "get_data"),
        ("developer", "add_template"),
        ("developer", "edit_template"),

        ('developer', "add_dimension"),
        ('developer', "update_dimension"),
        ('developer', "delete_dimension"),

        ('developer', "add_unit"),
        ('developer', "update_unit"),
        ('developer', "delete_unit"),

        # modeller permissions
        ("modeller", "add_network"),
        ("modeller", "edit_network"),
        ("modeller", "delete_network"),
        ("modeller", "share_network"),
        ("modeller", "edit_topology"),
        ("modeller", "get_project"),
        ("modeller", "add_project"),
        ("modeller", "edit_project"),
        ("modeller", "delete_project"),
        ("modeller", "share_project"),
        ("modeller", "edit_data"),
        ("modeller", "get_data"),

        # Manager permissions
        ("manager", "edit_data"),
        ("manager", "get_data"),
    )

    # Map for code to ID
    id_maps_dict = {
        "perm": {},
        "role": {}
    }
    # Adding perms
    perm_dict = {}
    for code, name in default_perms:
        perm = Perm(code=code, name=name)
        perm_dict[code] = perm
        perms_by_name = db.DBSession.query(Perm).filter(Perm.code==code).all()
        if len(perms_by_name)==0:
            # Adding perm
            log.debug("# Adding PERM {}".format(code))
            db.DBSession.add(perm)
            db.DBSession.flush()

        perm_by_name = db.DBSession.query(Perm).filter(Perm.code==code).one()
        id_maps_dict["perm"][code] = perm_by_name.id

    # Adding roles
    role_dict = {}
    for code, name in default_roles:
        role = Role(code=code, name=name)
        role_dict[code] = role
        roles_by_name = db.DBSession.query(Role).filter(Role.code==code).all()
        if len(roles_by_name)==0:
            # Adding perm
            log.debug("# Adding ROLE {}".format(code))
            db.DBSession.add(role)
            db.DBSession.flush()

        role_by_name = db.DBSession.query(Role).filter(Role.code==code).one()
        id_maps_dict["role"][code] = role_by_name.id

    # Adding connections
    for role_code, perm_code in roleperms:
        #log.info("Link Role:{}({}) <---> Perm:{}({})".format(role_code, id_maps_dict["role"][role_code], perm_code, id_maps_dict["perm"][perm_code]))

        links_found = db.DBSession.query(RolePerm).filter(RolePerm.role_id==id_maps_dict["role"][role_code]).filter(RolePerm.perm_id==id_maps_dict["perm"][perm_code]).all()
        if len(links_found) == 0:
            # Adding link
            log.debug("# Adding link")
            roleperm = RolePerm()
            # roleperm.role = role_dict[role_code]
            # roleperm.perm = perm_dict[perm_code]
            roleperm.role_id = id_maps_dict["role"][role_code]
            roleperm.perm_id = id_maps_dict["perm"][perm_code]
            db.DBSession.add(roleperm)
            db.DBSession.flush()

    db.DBSession.flush()

def create_default_units_and_dimensions(update=True):
    """
        Adds the units and the dimensions reading a json file. It adds only dimensions and units that are not inside the db
        It is possible adding new dimensions and units to the DB just modifiyin the json file

        args:
            update (bool) default True: If there are existing units / dimensions in the DB, then update
            them with the defaults by adding any missing units. If False, do nothing if there are existing
            units.
    """

    log.info("Adding default units and dimensions.")

    if update is False:
        #if update is set to false, check if there are dimensions. If there are
        #any existing dimensions, then log it and return.
        num_dimensions = db.DBSession.query(func.count(Dimension.id)).scalar()
        if num_dimensions > 0:
            log.info("Existing dimensions found. Not creating defaults.")
            return


    default_units_file_location = os.path.realpath(\
        os.path.join(os.path.dirname(os.path.realpath(__file__)),
                     '../',
                     'static',
                     'default_units_and_dimensions.json'))


    d = None

    with open(default_units_file_location) as json_data:
        d = json.load(json_data)
        json_data.close()

    for json_dimension in d["dimension"]:
        new_dimension = None
        dimension_name = get_utf8_encoded_string(json_dimension["name"])

        db_dimensions_by_name = db.DBSession.query(Dimension).filter(Dimension.name==dimension_name).all()

        if len(db_dimensions_by_name) == 0:
            # Adding the dimension
            log.debug("Adding Dimension `{}`".format(dimension_name))
            new_dimension = Dimension()
            if "id" in json_dimension:
                # If ID is specified
                new_dimension.id = json_dimension["id"]

            new_dimension.name = dimension_name

            db.DBSession.add(new_dimension)
            db.DBSession.flush()

        # Get the dimension by name
        new_dimension = get_dimension_from_db_by_name(dimension_name)

        for json_unit in json_dimension["unit"]:
            db_units_by_name = db.DBSession.query(Unit).filter(Unit.abbreviation==get_utf8_encoded_string(json_unit['abbr'])).all()
            if len(db_units_by_name) == 0:
                # Adding the unit
                log.debug("Adding Unit %s in %s",json_unit['abbr'], json_dimension["name"])
                new_unit = Unit()
                if "id" in json_unit:
                    new_unit.id = json_unit["id"]
                new_unit.dimension_id   = new_dimension.id
                new_unit.name           = get_utf8_encoded_string(json_unit['name'])
                new_unit.abbreviation   = get_utf8_encoded_string(json_unit['abbr'])
                new_unit.lf             = get_utf8_encoded_string(json_unit['lf'])
                new_unit.cf             = get_utf8_encoded_string(json_unit['cf'])
                if "description" in json_unit:
                    # If Description is specified
                    new_unit.description = get_utf8_encoded_string(json_unit["description"])

                # Save on DB
                db.DBSession.add(new_unit)
                db.DBSession.flush()
            else:
                #log.critical("UNIT {}.{} EXISTANT".format(dimension_name,json_unit['abbr']))
                pass
    try:
        # Needed for test. on HWI it fails so we need to catch the exception and pass by
        db.DBSession.commit()
    except Exception as e:
        # Needed for HWI
        pass
    return


def get_utf8_encoded_string(string):
    try:
        return string.encode('utf-8').strip().replace('"','\\"')
    except Exception as e:
        return string

def get_dimension_from_db_by_name(dimension_name):
    """
        Gets a dimension from the DB table.
    """
    try:
        dimension = db.DBSession.query(Dimension).filter(Dimension.name==dimension_name).one()
        return JSONObject(dimension)
    except NoResultFound:
        raise ResourceNotFoundError("Dimension %s not found"%(dimension_name))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

from beaker import session
import sqlalchemy as sa

from .. import util
from .. import config
from .. import db

DEFAULT_VALIDATE_KEY = 'YxaDbzUUSo08b+'
DEFAULT_DATA_DIR = '/tmp'
DEFAULT_FILE_DIR = '/tmp/auth'


CACHE_TABLE = sa.Table('beaker_cache', db.DeclarativeBase.metadata,
                       sa.Column('namespace', sa.String(255), primary_key=True),
                       sa.Column('accessed', sa.DateTime, nullable=False),
                       sa.Column('created', sa.DateTime, nullable=False),
                       sa.Column('data', sa.PickleType, nullable=False),
                       schema=db.DeclarativeBase.metadata.schema)

def login(username, password, **kwargs):
    """
        Login a user, returning a dict containing their user_id and session_id

        This does the DB login to check the credentials, and then creates a session
        so that requests from apps do not need to perform a login

        args:
            username (string): The user's username
            password(string): The user's password (unencrypted)
        returns:
            A dict containing the user_id and session_id
        raises:
            HydraError if the login fails
    """

    user_id = util.hdb.login_user(username, password)

    hydra_session = session.Session(
        {}, #This is normally a request object, but in this case is empty
        validate_key=config.get('COOKIES', 'VALIDATE_KEY', DEFAULT_VALIDATE_KEY),
        type='file' if db.hydra_db_url.startswith('sqlite') else 'ext:sqla',
        cookie_expires=True,
        data_dir=config.get('COOKIES', 'DATA_DIR', DEFAULT_DATA_DIR),
        bind=db.engine,
        table=CACHE_TABLE
    )

    hydra_session['user_id'] = user_id
    hydra_session['username'] = username
    hydra_session.save()

    return (user_id, hydra_session.id)

def logout(session_id, **kwargs):
    """
        Logout a user, removing their cookie if it exists and returning 'OK'


        args:
            session_id (string): The session ID to identify the cookie to remove
        returns:
            'OK'
        raises:
            HydraError if the logout fails
    """

    hydra_session_object = session.SessionObject(
        {}, #This is normally a request object, but in this case is empty
        validate_key=config.get('COOKIES', 'VALIDATE_KEY', DEFAULT_VALIDATE_KEY),
        type='file' if db.hydra_db_url.startswith('sqlite') else 'ext:sqla',
        cookie_expires=True,
        data_dir=config.get('COOKIES', 'DATA_DIR', DEFAULT_DATA_DIR),
        bind=db.engine,
        table=CACHE_TABLE
    )

    hydra_session = hydra_session_object.get_by_id(session_id)

    if hydra_session is not None:
        hydra_session.delete()
        hydra_session.namespace.do_remove()

    return 'OK'

def get_session_user(session_id, **kwargs):
    """
        Given a session ID, get the user ID that it is associated with

        args:
            session_id (string): The user's ID to identify the cookie to remove

        returns:
            user_id (string) or None if the session does not exist
    """

    hydra_session_object = session.SessionObject(
        {}, #This is normally a request object, but in this case is empty
        validate_key=config.get('COOKIES', 'VALIDATE_KEY', DEFAULT_VALIDATE_KEY),
        type='file' if db.hydra_db_url.startswith('sqlite') else 'ext:sqla',
        cookie_expires=True,
        data_dir=config.get('COOKIES', 'DATA_DIR', DEFAULT_DATA_DIR),
        bind=db.engine,
        table=CACHE_TABLE
    )

    hydra_session = hydra_session_object.get_by_id(session_id)

    if hydra_session is not None:
        return hydra_session

    return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#
#!/usr/bin/env python
# -*- coding: utf-8 -*-

try:
    from setuptools_scm import get_version
    __version__ = get_version(root='..', relative_to=__file__)
except ImportError:
    # setuptools_scm not available, fall back to hardcoded version
    __version__ = "0.1.14"
except Exception:
    # setuptools_scm failed to determine version, fall back to hardcoded version
    __version__ = "0.1.14"


import logging
from . import config
if config.CONFIG is None:
    config.load_config()


from . import hydra_logging
hydra_logging.init()

from .db import connect, commit_transaction, rollback_transaction

log = logging.getLogger(__name__)

log.debug(" \n ")

log.debug("CONFIG localfiles %s found in %s", len(config.localfiles), config.localfile)

log.debug("CONFIG repofiles %s found in %s", len(config.repofiles), config.repofile)

log.debug("CONFIG userfiles %s found in %s", len(config.userfiles), config.userfile)

log.debug("CONFIG sysfiles %s found in %s", len(config.sysfiles), config.sysfile)

if len(config.sysfiles) + len(config.repofiles) + len(config.userfiles) + len(config.sysfiles) == 0:
    log.critical("No config found. Please put your ini file into one of the files listed beside CONFIG above.")

if config.get("security", "max_login_attempts") is None:
    """  Absence of max_login_attempts results in all users unable to log in,
         so ensure this is defined in config, or fail.
    """
    raise RuntimeError("Config 'security' section must define 'max_login_attempts'")

log.debug(" \n ")

from .lib.attributes import *
from .lib.data import *
from .lib.groups import *
from .lib.network import *
from .lib.notes import *
from .lib.objects import *
from .lib.plugins import *
from .lib.project import *
from .lib.rules import *
from .lib.scenario import *
from .lib.sharing import *
from .lib.static import *
from .lib.template import *
from .lib.units import *
from .lib.users import *
from .lib.service import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#
from ..base import *

__all__ = ['ResourceAttr', 'ResourceAttrMap']

class ResourceAttr(Base, Inspect):
    """
    """

    __tablename__ = 'tResourceAttr'

    __table_args__ = (
        UniqueConstraint('network_id', 'attr_id', name='net_attr_1'),
        UniqueConstraint('project_id', 'attr_id', name='proj_attr_1'),
        UniqueConstraint('node_id', 'attr_id', name='node_attr_1'),
        UniqueConstraint('link_id', 'attr_id', name='link_attr_1'),
        UniqueConstraint('group_id', 'attr_id', name='group_attr_1'),
    )

    id = Column(Integer(), primary_key=True, nullable=False)
    attr_id = Column(Integer(), ForeignKey('tAttr.id'), nullable=False)
    ref_key = Column(String(60), nullable=False, index=True)
    network_id = Column(Integer(), ForeignKey('tNetwork.id'), index=True, nullable=True)
    project_id = Column(Integer(), ForeignKey('tProject.id'), index=True, nullable=True)
    node_id = Column(Integer(), ForeignKey('tNode.id'), index=True, nullable=True)
    link_id = Column(Integer(), ForeignKey('tLink.id'), index=True, nullable=True)
    group_id = Column(Integer(), ForeignKey('tResourceGroup.id'), index=True, nullable=True)
    attr_is_var = Column(String(1), nullable=False, server_default=text(u"'N'"))
    cr_date = Column(TIMESTAMP(), nullable=False, server_default=text(u'CURRENT_TIMESTAMP'))

    attr = relationship('Attr')
    project = relationship('Project',
                           backref=backref('attributes',
                                           uselist=True,
                                           cascade="all, delete-orphan"),
                           uselist=False)
    network = relationship('Network',
                           backref=backref('attributes',
                                           uselist=True,
                                           cascade="all, delete-orphan"),
                           uselist=False)
    node = relationship('Node',
                        backref=backref('attributes',
                                        uselist=True,
                                        cascade="all, delete-orphan"),
                        uselist=False)
    link = relationship('Link',
                        backref=backref('attributes',
                                        uselist=True,
                                        cascade="all, delete-orphan"),
                        uselist=False)
    resourcegroup = relationship('ResourceGroup',
                                 backref=backref('attributes',
                                                 uselist=True,
                                                 cascade="all, delete-orphan"),
                                 uselist=False)

    _parents = ['tNode', 'tLink', 'tResourceGroup', 'tNetwork', 'tProject']
    _children = []

    def get_network(self):
        """
         Get the network that this resource attribute is in.
        """
        ref_key = self.ref_key
        if ref_key == 'NETWORK':
            return self.network
        elif ref_key == 'NODE':
            return self.node.network
        elif ref_key == 'LINK':
            return self.link.network
        elif ref_key == 'GROUP':
            return self.group.network
        elif ref_key == 'PROJECT':
            return None

    def get_resource(self):
        ref_key = self.ref_key
        if ref_key == 'NETWORK':
            return self.network
        elif ref_key == 'NODE':
            return self.node
        elif ref_key == 'LINK':
            return self.link
        elif ref_key == 'GROUP':
            return self.resourcegroup
        elif ref_key == 'PROJECT':
            return self.project

    def get_resource_id(self):
        ref_key = self.ref_key
        if ref_key == 'NETWORK':
            return self.network_id
        elif ref_key == 'NODE':
            return self.node_id
        elif ref_key == 'LINK':
            return self.link_id
        elif ref_key == 'GROUP':
            return self.group_id
        elif ref_key == 'PROJECT':
            return self.project_id

    def check_read_permission(self, user_id, do_raise=True, is_admin=None):
        """
            Check whether this user can read this resource attribute
        """
        return self.get_resource().check_read_permission(user_id, do_raise=do_raise, is_admin=is_admin)

    def check_write_permission(self, user_id, do_raise=True, is_admin=None):
        """
            Check whether this user can write this node
        """
        return self.get_resource().check_write_permission(user_id, do_raise=do_raise, is_admin=is_admin)

class ResourceAttrMap(Base, Inspect):
    """
    """

    __tablename__ = 'tResourceAttrMap'

    network_a_id = Column(Integer(), ForeignKey('tNetwork.id'),
                          primary_key=True, nullable=False)
    network_b_id = Column(Integer(), ForeignKey('tNetwork.id'),
                          primary_key=True, nullable=False)
    resource_attr_id_a = Column(Integer(), ForeignKey('tResourceAttr.id'),
                                primary_key=True, nullable=False)
    resource_attr_id_b = Column(Integer(), ForeignKey('tResourceAttr.id'),
                                primary_key=True, nullable=False)

    resourceattr_a = relationship("ResourceAttr", foreign_keys=[resource_attr_id_a])
    resourceattr_b = relationship("ResourceAttr", foreign_keys=[resource_attr_id_b])

    network_a = relationship("Network", foreign_keys=[network_a_id])
    network_b = relationship("Network", foreign_keys=[network_b_id])


# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#
#!/usr/bin/env python
# -*- coding: utf-8 -*-

from sqlalchemy import create_engine,\
        MetaData,\
        Table,\
        Column,\
        Integer,\
        String,\
        TIMESTAMP,\
        text,\
        DDL
from sqlalchemy.engine import reflection
import logging
from hydra_base import config
from subprocess import Popen
from sqlalchemy.types import DECIMAL, NUMERIC
from sqlalchemy.dialects.mysql.base import DOUBLE
from decimal import Decimal
import os

engine_name = config.get('mysqld', 'url')
sqlite_engine = "sqlite:///%s"%(config.get('sqlite', 'backup_url'))

def connect():
    """
        return an inspector object
    """
    db = create_engine(engine_name, echo=True)
    db.connect()

    return db

def create_sqlite_backup_db(audit_tables):
    """
        return an inspector object
    """
    #we always want to create a whole new DB, so delete the old one first
    #if it exists.
    try:
        Popen("rm %s"%(config.get('sqlite', 'backup_url')), shell=True)
        logging.warn("Old sqlite backup DB removed")
    except Exception as e:
        logging.warn(e)

    try:
        aux_dir = config.get('DEFAULT', 'hydra_aux_dir')
        os.mkdir(aux_dir)
        logging.warn("%s created", aux_dir)
    except Exception as e:
        logging.warn(e)

    try:
        backup_dir = config.get('db', 'export_target')
        os.mkdir(backup_dir)
        logging.warn("%s created", backup_dir)
    except Exception as e:
        logging.warn(e)

    db = create_engine(sqlite_engine, echo=True)
    db.connect()
    metadata = MetaData(db)

    for main_audit_table in audit_tables:
        cols = []
        for c in main_audit_table.columns:
            col = c.copy()
            if col.type.python_type == Decimal:
                col.type = DECIMAL()

            cols.append(col)
        Table(main_audit_table.name, metadata, *cols, sqlite_autoincrement=True)

    metadata.create_all(db)

def run():
    db = create_engine(engine_name, echo=True)
    db = connect()
    metadata = MetaData(db)
    insp = reflection.Inspector.from_engine(db)
    tables = []
    for table_name in insp.get_table_names():
        table = Table(table_name, metadata, autoload=True, autoload_with=db)
        if not table_name.endswith('_aud'):
            tables.append(table)
        else:
            table.drop(db)
            metadata.remove(table)

    audit_tables = []
    for t in tables:
        audit_table = create_audit_table(t)
        audit_tables.append(audit_table)

    #create_sqlite_backup_db(audit_tables)
    create_triggers(db, tables)
    metadata.create_all()

def create_triggers(db, tables):


    db = create_engine(engine_name)
    db.echo = True
    db.connect()
    metadata = MetaData(db)


    insp = reflection.Inspector.from_engine(db)

    tables = []
    for table_name in insp.get_table_names():
        if not table_name.endswith('_aud'):
            table = Table(table_name, metadata, autoload=True, autoload_with=db)
            tables.append(table)
            #print("TABLE: %s"%table)
            #print table.__repr__
        else:
            table = Table(table_name, metadata, autoload=True, autoload_with=db)
            table.drop(db)
            metadata.remove(table)


    drop_trigger_text = """DROP TRIGGER IF EXISTS %(trigger_name)s;"""
    for table in tables:
        pk_cols = [c.name for c in table.primary_key]
        for pk_col in pk_cols:
            try:
                db.execute(drop_trigger_text % {
                    'trigger_name' : table.name + "_ins_trig",
                })
            except:
                pass

        for pk_col in pk_cols:
            try:
                db.execute(drop_trigger_text % {
                    'trigger_name' : table.name + "_upd_trig",
                })
            except:
                pass
    #metadata.create_all()

    trigger_text = """
                    CREATE TRIGGER
                        %(trigger_name)s
                    AFTER %(action)s ON
                        %(table_name)s
                    FOR EACH ROW
                        BEGIN
                            INSERT INTO %(table_name)s_aud
                            SELECT
                                d.*,
                                '%(action)s',
                                NULL,
                                NOW()
                            FROM
                                %(table_name)s
                                AS d
                            WHERE
                                %(pkd)s;
                        END
                        """

    for table in tables:


        pk_cols = [c.name for c in table.primary_key]
        pkd = []

        for pk_col in pk_cols:
            pkd.append("d.%s = NEW.%s"%(pk_col, pk_col))

        text_dict = {
            'action'       : 'INSERT',
            'trigger_name' : table.name + "_ins_trig",
            'table_name'   : table.name,
            'pkd'           : ' and '.join(pkd),
        }

        logging.info(trigger_text % text_dict)
        trig_ddl = DDL(trigger_text % text_dict)
        trig_ddl.execute_at('after-create', table.metadata)

        text_dict['action'] = 'UPDATE'
        text_dict['trigger_name'] = table.name + "_upd_trig"
        trig_ddl = DDL(trigger_text % text_dict)
        trig_ddl.execute_at('after-create', table.metadata)

    metadata.create_all()

def create_audit_table(table):
    args = []
    for c in table.columns:
        col = c.copy()
        #if col.name == 'cr_date':
        #    continue
        col.primary_key=False
        col.foreign_keys = set([])
        col.server_default=None
        col.default=None
        col.nullable=True
        args.append(col)
    args.append(Column('action', String(12)))
    args.append(Column('aud_id', Integer, primary_key=True))
#    args.append(Column('aud_user_id', Integer, ForeignKey('tUser.id')))
    args.append(Column('aud_time', TIMESTAMP, server_default=text('LOCALTIMESTAMP')))
    return Table(table.name+"_aud", table.metadata, *args, extend_existing=True)

if __name__ == '__main__':
    logging.basicConfig(level='INFO')
    run()
   # create_triggers()



"""resourcegroupitem

Revision ID: 6b24d6443e31
Revises: f52ccd2b7ffb
Create Date: 2018-04-17 15:34:25.697772

"""
from alembic import op
import sqlalchemy as sa

import logging
log = logging.getLogger(__name__)

# revision identifiers, used by Alembic.
revision = '6b24d6443e31'
down_revision = 'f52ccd2b7ffb'
branch_labels = None
depends_on = None


def upgrade():
    if op.get_bind().dialect.name == 'mysql':

        # ### tResourceGroupItemt

        try:
            op.alter_column('tResourceGroupItem', 'item_id', new_column_name='id', existing_type=sa.Integer(), primary_key=True, autoincrement=True, nullable=False)
        except Exception as e:
            log.exception(e)

    else: ## sqlite

        # ## tResourceGroupItem
        try:
            op.drop_table('tResourceGroupItem_new')
        except:
            log.info("tResourceGroupItem_new isn't there")

        try:
            # ## tResourceGroupItem
            op.create_table(
                'tResourceGroupItem_new',
                    sa.Column('id', sa.Integer(), primary_key=True, nullable=False),
                    sa.Column('group_id', sa.Integer(), sa.ForeignKey('tResourceGroup.id')),
                    sa.Column('scenario_id', sa.Integer(), sa.ForeignKey('tScenario.id'),  nullable=False, index=True),
                    sa.Column('ref_key', sa.String(60),  nullable=False),
                    sa.Column('node_id', sa.Integer(),  sa.ForeignKey('tNode.id')),
                    sa.Column('link_id', sa.Integer(),  sa.ForeignKey('tLink.id')),
                    sa.Column('subgroup_id', sa.Integer(),  sa.ForeignKey('tResourceGroup.id')),
                    sa.Column('cr_date', sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
                    sa.UniqueConstraint('group_id', 'node_id', 'scenario_id', name='node_group_1'),
                    sa.UniqueConstraint('group_id', 'link_id', 'scenario_id',  name = 'link_group_1'),
                    sa.UniqueConstraint('group_id', 'subgroup_id', 'scenario_id', name = 'subgroup_group_1'),
            )

            op.execute("insert into tResourceGroupItem_new (id, group_id, scenario_id, ref_key, node_id, link_id, subgroup_id, cr_date) select item_id, group_id, scenario_id, ref_key, node_id, link_id, subgroup_id, cr_date from tResourceGroupItem")

            op.rename_table('tResourceGroupItem','tResourceGroupItem_old')
            op.rename_table('tResourceGroupItem_new', 'tResourceGroupItem')
            op.drop_table('tResourceGroupItem_old')

        except Exception as e:
            log.exception(e)


def downgrade():
    if op.get_bind().dialect.name == 'mysql':

        # ### tResourceGroupItemt

        try:
            op.alter_column('tResourceGroupItem', 'id', new_column_name='item_id', existing_type=sa.Integer(), primary_key=True, autoincrement=True, nullable=False)
        except Exception as e:
            log.exception(e)

    else: ## sqlite

        # ## tResourceGroupItem
        try:
            op.drop_table('tResourceGroupItem_new')
        except:
            log.info("tResourceGroupItem_new isn't there")

        try:
            # ## tResourceGroupItem
            op.create_table(
                'tResourceGroupItem_new',
                    sa.Column('item_id', sa.Integer(), primary_key=True, nullable=False),
                    sa.Column('group_id', sa.Integer(), sa.ForeignKey('tResourceGroup.id')),
                    sa.Column('scenario_id', sa.Integer(), sa.ForeignKey('tScenario.id'),  nullable=False, index=True),
                    sa.Column('ref_key', sa.String(60),  nullable=False),
                    sa.Column('node_id', sa.Integer(),  sa.ForeignKey('tNode.id')),
                    sa.Column('link_id', sa.Integer(),  sa.ForeignKey('tLink.id')),
                    sa.Column('subgroup_id', sa.Integer(),  sa.ForeignKey('tResourceGroup.id')),
                    sa.Column('cr_date', sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
                    sa.UniqueConstraint('group_id', 'node_id', 'scenario_id', name='node_group_1'),
                    sa.UniqueConstraint('group_id', 'link_id', 'scenario_id',  name = 'link_group_1'),
                    sa.UniqueConstraint('group_id', 'subgroup_id', 'scenario_id', name = 'subgroup_group_1'),
            )

            op.execute("insert into tResourceGroupItem_new (item_id, group_id, scenario_id, ref_key, node_id, link_id, subgroup_id, cr_date) select id, group_id, scenario_id, ref_key, node_id, link_id, subgroup_id, cr_date from tResourceGroupItem")

            op.rename_table('tResourceGroupItem','tResourceGroupItem_old')
            op.rename_table('tResourceGroupItem_new', 'tResourceGroupItem')
            op.drop_table('tResourceGroupItem_old')

        except Exception as e:
            log.exception(e)



"""resource_attr_id

Revision ID: 08f0ebe40290
Revises: a7896842f484
Create Date: 2018-04-16 14:49:24.202974

"""
from alembic import op
import sqlalchemy as sa

import logging
log = logging.getLogger(__name__)

# revision identifiers, used by Alembic.
revision = '08f0ebe40290'
down_revision = 'a7896842f484'
branch_labels = None
depends_on = None


def upgrade():
    if op.get_bind().dialect.name == 'mysql':

        # ### tResourceAttrt

        try:
            op.alter_column('tResourceAttr', 'resource_attr_id', new_column_name='id', existing_type=sa.Integer(), primary_key=True, autoincrement=True, nullable=False)
        except Exception as e:
            log.exception(e)

    else: ## sqlite

        # ## tResourceAttr
        try:
            op.drop_table('tResourceAttr_new')
        except:
            log.info("tResourceAttr_new isn't there")

        try:
            # ## tResourceAttr
            op.create_table(
                'tResourceAttr_new',
                sa.Column('id', sa.Integer(), primary_key=True, nullable=False),
                sa.Column('attr_id', sa.Integer(), sa.ForeignKey('tAttr.id'),  nullable=False),
                sa.Column('ref_key', sa.String(60),  nullable=False, index=True),
                sa.Column('network_id', sa.Integer(),  sa.ForeignKey('tNetwork.id'), index=True, nullable=True,),
                sa.Column('project_id', sa.Integer(),  sa.ForeignKey('tProject.id'), index=True, nullable=True,),
                sa.Column('node_id', sa.Integer(),  sa.ForeignKey('tNode.id'), index=True, nullable=True),
                sa.Column('link_id', sa.Integer(),  sa.ForeignKey('tLink.id'), index=True, nullable=True),
                sa.Column('group_id', sa.Integer(),  sa.ForeignKey('tResourceGroup.id'), index=True, nullable=True),
                sa.Column('attr_is_var', sa.String(1),  nullable=False, server_default=sa.text(u"'N'")),
                sa.Column('cr_date', sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
                sa.UniqueConstraint('network_id', 'attr_id', name = 'net_attr_1'),
                sa.UniqueConstraint('project_id', 'attr_id', name = 'proj_attr_1'),
                sa.UniqueConstraint('node_id',    'attr_id', name = 'node_attr_1'),
                sa.UniqueConstraint('link_id',    'attr_id', name = 'link_attr_1'),
                sa.UniqueConstraint('group_id',   'attr_id', name = 'group_attr_1'),
            )

            op.execute("insert into tResourceAttr_new (id, attr_id, ref_key, network_id, project_id, node_id, link_id, group_id, attr_is_var, cr_date) select resource_attr_id, attr_id, ref_key, network_id, project_id, node_id, link_id, group_id, attr_is_var, cr_date from tResourceAttr")

            op.rename_table('tResourceAttr','tResourceAttr_old')
            op.rename_table('tResourceAttr_new', 'tResourceAttr')
            op.drop_table('tResourceAttr_old')

        except Exception as e:
            log.exception(e)


def downgrade():
    if op.get_bind().dialect.name == 'mysql':

        # ### tProject

        try:
            op.alter_column('tResourceAttr', 'id', new_column_name='resource_attr_id', existing_type=sa.Integer(), primary_key=True, autoincrement=True, nullable=False)
        except Exception as e:
            log.exception(e)

    else: ## sqlite

        # ## tResourceAttr
        try:
            op.drop_table('tResourceAttr_new')
        except:
            log.info("tResourceAttr_new isn't there")

        try:
            # ## tResourceAttr
            op.create_table(
                'tResourceAttr_new',
                sa.Column('resource_attr_id', sa.Integer(), primary_key=True, nullable=False),
                sa.Column('attr_id', sa.Integer(), sa.ForeignKey('tAttr.id'),  nullable=False),
                sa.Column('ref_key', sa.String(60),  nullable=False, index=True),
                sa.Column('network_id', sa.Integer(),  sa.ForeignKey('tNetwork.id'), index=True, nullable=True,),
                sa.Column('project_id', sa.Integer(),  sa.ForeignKey('tProject.id'), index=True, nullable=True,),
                sa.Column('node_id', sa.Integer(),  sa.ForeignKey('tNode.id'), index=True, nullable=True),
                sa.Column('link_id', sa.Integer(),  sa.ForeignKey('tLink.id'), index=True, nullable=True),
                sa.Column('group_id', sa.Integer(),  sa.ForeignKey('tResourceGroup.id'), index=True, nullable=True),
                sa.Column('attr_is_var', sa.String(1),  nullable=False, server_default=sa.text(u"'N'")),
                sa.Column('cr_date', sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
                sa.UniqueConstraint('network_id', 'attr_id', name = 'net_attr_1'),
                sa.UniqueConstraint('project_id', 'attr_id', name = 'proj_attr_1'),
                sa.UniqueConstraint('node_id',    'attr_id', name = 'node_attr_1'),
                sa.UniqueConstraint('link_id',    'attr_id', name = 'link_attr_1'),
                sa.UniqueConstraint('group_id',   'attr_id', name = 'group_attr_1'),
            )

            op.execute("insert into tResourceAttr_new (resource_attr_id, attr_id, ref_key, network_id, project_id, node_id, link_id, group_id, attr_is_var, cr_date) select id, attr_id, ref_key, network_id, project_id, node_id, link_id, group_id, attr_is_var, cr_date from tResourceAttr")

            op.rename_table('tResourceAttr','tResourceAttr_old')
            op.rename_table('tResourceAttr_new', 'tResourceAttr')
            op.drop_table('tResourceAttr_old')

        except Exception as e:
            log.exception(e)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

import hydra_base
from hydra_base.exceptions import HydraError
from hydra_base.lib.objects import JSONObject
import datetime
import bcrypt
import pytest


@pytest.fixture
def user_json_object():
    user = JSONObject(dict(
        username="test_user @ %s" % (datetime.datetime.now()),
        password="test_user_password",
        display_name="Tom"
    ))
    return user


@pytest.fixture
def role_json_object():
    role = JSONObject(dict(
        name="Test Role",
        code="test_role @ %s" % (datetime.datetime.now())
    ))
    return role


@pytest.fixture
def perm_json_object():
    perm = JSONObject(dict(
        name="Test Perm",
        code="test_perm @ %s" % (datetime.datetime.now())
    ))
    return perm


class TestUser:
    """ A collection of tests of the User part of the DB.
    """

    def test_add_user(self, client, user_json_object):
        """ Test adding a new user to the DB """
        user = user_json_object  # Convenience renaming

        new_user = client.add_user(user)
        assert new_user.username == user.username, "Usernames are not the same!"
        #convert pwd to bytes if it's not bytes
        #Add user returns the password *hash*, not the plain-text password
        pwd_hash = new_user.password if isinstance(new_user.password, bytes) else new_user.password.encode('utf-8')
        assert bcrypt.checkpw(user.password.encode('utf-8'), pwd_hash)
        assert new_user.display_name == user.display_name

        new_user.display_name = 'Tom Update'

        updated_user = client.update_user_display_name(new_user)

        assert updated_user.display_name == new_user.display_name

        delete_result = client.delete_user(new_user.id)

        assert delete_result == 'OK', "User was not removed!"

    def test_update_user_password(self, client, user_json_object):
        """ Test changing a user's password """
        user = user_json_object

        new_user = client.add_user(user)

        # TODO create a new session here
        new_password = 'new_test_user_password'
        new_user = client.update_user_password(new_user.id, new_password)
        pwd = new_user.password
        pwd_hash = pwd if isinstance(pwd, bytes) else pwd.encode('utf-8')

        assert bcrypt.checkpw(new_password.encode('utf-8'), pwd_hash)

    def test_add_role(self, client, role_json_object):
        """ Test adding a new role """
        role = role_json_object
        new_role = client.add_role(role)

        assert new_role.id is not None, "Role does not have an ID!"
        assert new_role.name == role.name, "Role are not the same!"

        delete_result = client.delete_role(new_role.id)

        assert delete_result == 'OK', "Role was not removed!"

    def test_add_perm(self, client, perm_json_object):
        """ Test adding a new permission """
        perm = perm_json_object

        new_perm = client.add_perm(perm)

        assert new_perm.id is not None, "Perm does not have an ID!"
        assert new_perm.name == perm.name, "Perm are not the same!"

        delete_result = client.delete_perm(new_perm.id)

        assert delete_result == 'OK', "perm was not removed!"

    def test_set_user_role(self, client, user_json_object, role_json_object):
        """ Test assigning a role to a user """
        # Rename for convenience during test
        user = user_json_object
        role = role_json_object

        # Add the user and the role
        new_user = client.add_user(user)
        new_role = client.add_role(role)

        # Apply role to the user
        # NB `set_user_role` returns a `Role` instance not a `RoleUser` instance
        role_with_users = client.set_user_role(new_user.id, new_role.id)

        assert role_with_users is not None, "Role user was not set correctly"
        assert role_with_users.roleusers[0].user_id == new_user.id, "User was not added to role correctly."

        delete_result = client.delete_user_role(new_user.id, new_role.id)

        assert delete_result == 'OK', "Role User was not removed!"

        delete_result = client.delete_user(new_user.id)

        assert delete_result == 'OK', "Role User was not removed!"

    def test_set_role_perm(self, client, role_json_object, perm_json_object):
        """ Test assigning a perm to a user """
        # Rename for convenience during testing
        role = role_json_object
        perm = perm_json_object

        # Add the role and perm
        new_role = client.add_role(role)
        new_perm = client.add_perm(perm)

        role_with_perms = client.set_role_perm(new_role.id, new_perm.id)

        assert role_with_perms is not None, "Role perm was not set correctly"
        assert role_with_perms.roleperms[0].perm_id == new_perm.id, "Perm was not added to role correctly."

        delete_result = client.delete_role_perm(new_role.id, new_perm.id)

        assert delete_result == 'OK', "Role Perm was not removed!"

        delete_result = client.delete_perm(new_perm.id)

        assert delete_result == 'OK', "Role Perm was not removed!"

    def test_get_users(self, client, user_json_object):
        """ Test fetching all the users """

        # These are the users added for the tests.
        existing_users = {1: 'root', 2: 'UserA', 3: 'UserB', 4: 'UserC'}

        # Add one user
        client.add_user(user_json_object)
        # There should always be a root user
        users = client.get_all_users()
        assert len(users) >= 4
        assert set(existing_users.values()).issubset(set([u.username for u in users]))

    def test_get_username(self, client, user_json_object):
        """ Test retrieving a username """
        # Add one user
        client.add_user(user_json_object)
        username = client.get_username(1)
        assert username == 'root'

    def test_get_perms(self, client):
        """ Test fetching all the perms """
        perms = client.get_all_perms()
        assert len(perms) >= 19

        check_perm = perms[0]

        single_perm = client.get_perm(check_perm.id)

        assert single_perm.id == check_perm.id
        assert single_perm.code == check_perm.code
        assert single_perm.name == check_perm.name

        # Check invalid ID raises an error
        with pytest.raises(HydraError):
            client.get_perm(99999)

    def test_get_roles(self, client):
        roles = client.get_all_roles()
        assert len(roles) >= 6

        role_codes = set([r.code for r in roles])
        core_set = set(['admin', 'dev', 'modeller', 'manager', 'grad', 'decision'])
        assert core_set.issubset(role_codes)
        for r in roles:
            if r.code == 'admin':
                assert len(r.roleperms) >= 10
                check_role = r

        single_role = client.get_role(check_role.id)
        assert check_role.id == single_role.id
        assert check_role.code == single_role.code
        assert check_role.name == single_role.name
        assert len(check_role.roleperms) == len(single_role.roleperms)

        # Check invalid ID raises an error
        with pytest.raises(HydraError):
            client.get_role(99999)

    def test_get_user_roles(self, client):
        roles = client.get_user_roles(1)
        assert len(roles) == 1
        assert roles[0].code == 'admin'

    def test_get_user_permissions(self, client):
        permissions = client.get_user_permissions(1)

        role = client.get_role_by_code('admin')

        assert len(permissions) == len(role.roleperms)

    def test_user_has_failed_login_count(self, client, user_json_object):
        user = client.add_user(user_json_object)
        failed_logins = client.get_failed_login_count(user.username)

    def test_max_login_attempts_defined(self, client, user_json_object):
        user = client.add_user(user_json_object)
        max_attempts = client.get_max_login_attempts()
        assert max_attempts is not None

    def test_user_has_remaining_login_attempts(self, client, user_json_object):
        user = client.add_user(user_json_object)
        remaining_attempts = client.get_remaining_login_attempts(user.username)

    def test_inc_user_failed_logins(self, client, user_json_object):
        user = client.add_user(user_json_object)

        initial_failed_logins = client.get_failed_login_count(user.username)
        client.inc_failed_login_attempts(user.username)
        final_failed_logins = client.get_failed_login_count(user.username)

        assert final_failed_logins == initial_failed_logins + 1

    def test_reset_user_failed_logins(self, client, user_json_object):
        user = client.add_user(user_json_object)

        client.inc_failed_login_attempts(user.username)
        client.reset_failed_logins(user.username)
        failed_logins = client.get_failed_login_count(user.username)

        assert failed_logins == 0

    def test_login_count_totals(self, client, user_json_object):
        user = client.add_user(user_json_object)

        client.inc_failed_login_attempts(user.username)
        failed_logins = client.get_failed_login_count(user.username)
        remaining_attempts = client.get_remaining_login_attempts(user.username)
        max_attempts = client.get_max_login_attempts()

        assert max_attempts == failed_logins + remaining_attempts



"""attr_id

Revision ID: f579b06e9cc2
Revises: dd1110144e53
Create Date: 2018-04-11 21:28:54.173256

"""
from alembic import op
import sqlalchemy as sa

import logging
log = logging.getLogger(__name__)


# revision identifiers, used by Alembic.
revision = 'f579b06e9cc2'
down_revision = 'dd1110144e53'
branch_labels = None
depends_on = None


def upgrade():
    if op.get_bind().dialect.name == 'mysql':

        # ### tProject

        try:
            op.alter_column('tAttr', 'attr_id', new_column_name='id', existing_type=sa.Integer(), autoincrement=True, nullable=False)
            op.alter_column('tAttr', 'attr_name', new_column_name='name', existing_type=sa.String(60), nullable=False)
            op.alter_column('tAttr', 'attr_dimen', new_column_name='dimension', existing_type=sa.String(60), server_default=sa.text(u"'dimensionless'"))
            op.alter_column('tAttr', 'attr_description', new_column_name='description', existing_type=sa.String(1000))
        except Exception as e:
            log.exception(e)
    else: ## sqlite

        # ## tAttr
        try:
            op.drop_table('tAttr_new')
        except:
            log.info("tAttr_new isn't there")

        try:
            # ## tAttr
            op.create_table(
                'tAttr_new',
                sa.Column('id', sa.Integer, primary_key=True),
                sa.Column('name', sa.Text(60), nullable=False),
                sa.Column('dimension', sa.String(60),  nullable=False),
                sa.Column('description', sa.String(1000)),
                sa.Column('cr_date', sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
            )

            op.execute("insert into tAttr_new (id, name, dimension, description, cr_date) select attr_id, attr_name, attr_dimen, attr_description, cr_date from tAttr")

            op.rename_table('tAttr','tAttr_old')
            op.rename_table('tAttr_new', 'tAttr')
            op.drop_table('tAttr_old')

        except Exception as e:
            log.exception(e)

        # ## tMetadata
        try:
            op.drop_table('tMetadata_new')
        except:
            log.info("tMetadata_new isn't there")

        try:
            # ## tMetadata
            op.create_table(
                'tMetadata_new',
                sa.Column('dataset_id', sa.Integer, primary_key=True, index=True),
                sa.Column('key', sa.String(60), nullable=False, primary_key=True),
                sa.Column('value', sa.Text(),  nullable=False),
            )

            op.execute("insert into tMetadata_new (dataset_id, key, value) select dataset_id, metadata_name, metadata_val from tMetadata")

            op.rename_table('tMetadata','tMetadata_old')
            op.rename_table('tMetadata_new', 'tMetadata')
            op.drop_table('tMetadata_old')

        except Exception as e:
            log.exception(e)


def downgrade():
    if op.get_bind().dialect.name == 'mysql':

        # ### tProject

        try:
            op.alter_column('tAttr', 'id', new_column_name='attr_id', existing_type=sa.Integer(), autoincrement=True, nullable=False)
            op.alter_column('tAttr', 'name', new_column_name='attr_name', existing_type=sa.String(60), nullable=False)
            op.alter_column('tAttr', 'dimension', new_column_name='attr_dimen', existing_type=sa.String(60), server_default=text(u"'dimensionless'"))
            op.alter_column('tAttr', 'description', new_column_name='attr_description', existing_type=sa.String(1000))
        except Exception as e:
            log.exception(e)
    else: ## sqlite

        # ## tAttr
        try:
            op.drop_table('tAttr_new')
        except:
            log.info("tAttr_new isn't there")

        try:
            # ## tAttr
            op.create_table(
                'tAttr_new',
                sa.Column('attr_id', sa.Integer, primary_key=True),
                sa.Column('attr_name', sa.Text(60), nullable=False),
                sa.Column('attr_dimen', sa.String(60),  nullable=False),
                sa.Column('attr_description', sa.String(1000)),
                sa.Column('cr_date', sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
            )

            op.execute("insert into tAttr_new (attr_id, attr_name, attr_dimen, attr_description, cr_date) select id, name, dimension, description, cr_date from tAttr")

            op.rename_table('tAttr','tAttr_old')
            op.rename_table('tAttr_new', 'tAttr')
            op.drop_table('tAttr_old')

        except Exception as e:
            log.exception(e)

        # ## tMetadata
        try:
            op.drop_table('tMetadata_new')
        except:
            log.info("tMetadata_new isn't there")

        try:
            # ## tMetadata
            op.create_table(
                'tMetadata_new',
                sa.Column('dataset_id', sa.Integer, primary_key=True, index=True),
                sa.Column('key', sa.String(60), nullable=False, primary_key=True),
                sa.Column('value', sa.Text(),  nullable=False),
            )

            op.execute("insert into tMetadata_new (dataset_id, key, value) select dataset_id, metadata_name, metadata_val from tMetadata")

            op.rename_table('tMetadata','tMetadata_old')
            op.rename_table('tMetadata_new', 'tMetadata')
            op.drop_table('tMetadata_old')

        except Exception as e:
            log.exception(e)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#
from .base import *

__all__ = ['Template', 'TemplateType', 'TypeAttr', 'ResourceType']

from .attributes import Attr
from .units import Unit


class Template(Base, Inspect):
    """
    Template
    """

    __tablename__ = 'tTemplate'

    id = Column(Integer(), primary_key=True, nullable=False)
    parent_id = Column(Integer(), ForeignKey('tTemplate.id'))
    name = Column(String(200), unique=True)
    status = Column(String(1),  nullable=False, server_default=text(u"'A'"))
    description = Column(String(1000))
    cr_date = Column(TIMESTAMP(), nullable=False, server_default=text(u'CURRENT_TIMESTAMP'))
    layout = Column(Text().with_variant(mysql.LONGTEXT, 'mysql'))

    parent = relationship('Template', remote_side=[id], backref=backref("children", order_by=id))

    _parents = []
    _children = ['tTemplateType']

    def set_inherited_columns(self, parent, child, table):
        """
            Set the value on the column of a target child.
            This checks if the value is null on the child, and sets it from
            the parent if so
        """
        inherited_columns = []
        for column in table.__table__.columns:

            colname = column.name

            if hasattr(table, '_protected_columns')\
               and colname in table._protected_columns:
                # as a child, you can't change stuff like IDs, cr dates etc.
                continue

            newval = getattr(parent, colname)

            if colname == 'layout':
                newval = get_json_as_string(newval)

            refval = getattr(child, colname)

            if refval is None:
                inherited_columns.append(colname)
                setattr(child, colname, newval)

        if hasattr(child, 'inherited_columns') and child.inherited_columns is not None:
            for c in inherited_columns:
                if c not in child.inherited_columns:
                    child.inherited_columns.append(c)
        else:
            child.inherited_columns = inherited_columns

        return child

    def get_typeattr(self, typeattr_id, child_typeattr=None, get_parent_types=True):
        """
            Return an individual type attribute.
            If this type attribute inherits from another, look up the tree to compile
            the inherited data, with the leaf node taking priority
        """

        #get all the type attrs for this type, and add any which are missing
        this_typeattr_i = get_session().query(TypeAttr)\
            .filter(TypeAttr.id == typeattr_id)\
            .options(joinedload(TypeAttr.attr))\
            .options(joinedload(TypeAttr.default_dataset)).one()

        this_typeattr = JSONObject(this_typeattr_i)

        if child_typeattr is None:
            child_typeattr = this_typeattr
        else:
            child_typeattr = self.set_inherited_columns(this_typeattr, child_typeattr, this_typeattr_i)

        if this_typeattr.parent_id is not None and get_parent_types is True:
            return self.parent.get_typeattr(this_typeattr.parent_id,
                                            child_typeattr=child_typeattr,
                                            get_parent_types=get_parent_types)

        return child_typeattr

    def get_type(self, type_id, child_type=None, get_parent_types=True):
        """
            Return all the templatetypes relevant to this template.
            If this template inherits from another, look up the tree to compile
            an exhaustive list of types, removing any duplicates, prioritising
            the ones closest to this template (my immediate parent's values are used instead
            of its parents)

        """

        #Add resource attributes which are not defined already
        this_type_i = get_session().query(TemplateType).filter(
            TemplateType.id == type_id).options(noload(TemplateType.typeattrs)).one()

        this_type = JSONObject(this_type_i)

        this_type.typeattrs = []

        if not hasattr(this_type, 'ta_tree') or this_type.ta_tree is None:
            this_type.ta_tree = {}

        #get all the type attrs for this type, and add any which are missing
        typeattrs_i = get_session().query(TypeAttr)\
            .filter(TypeAttr.type_id == type_id)\
            .options(joinedload(TypeAttr.attr))\
            .options(joinedload(TypeAttr.unit))\
            .options(joinedload(TypeAttr.default_dataset)).all()
        typeattrs = [JSONObject(ta) for ta in typeattrs_i]


        #Is this type the parent of a type. If so, we don't want to add a new type
        #we want to update an existing one with any data that it's missing
        if child_type is not None:
            child_type = self.set_inherited_columns(this_type, child_type, this_type_i)
            #check the child's typeattrs. If a typeattr exists on the parent, with the
            #same attr_id, then it should be ignore. THis can happen when adding a typeattr
            #to the child first, then the parent
            child_typeattrs = [ta.attr_id for ta in child_type.typeattrs]

            for i, typeattr in enumerate(typeattrs):
                if typeattr.attr_id in child_typeattrs:
                    log.debug("Found a typeattr for attribute %s on the "
                             "child type %s (%s). Ignoring",
                             typeattr.attr_id, child_type.name, child_type.id)
                    continue
                #Does this typeattr have a child?
                child_typeattr = child_type.ta_tree.get(typeattr.id)
                if child_typeattr is None:
                    #there is no child, so check if it is a child
                    if typeattr.parent_id is not None:
                        #it has a parent, so add it to the type's tree dict
                        #for processing farther up the tree
                        child_type.ta_tree[typeattr.parent_id] = typeattr
                    child_type.typeattrs.append(typeattr)
                else:
                    child_typeattr = self.set_inherited_columns(typeattr, child_typeattr, typeattrs_i[i])

        else:
            if not hasattr(this_type, 'typeattrs'):
                setattr(this_type, 'typeattrs', [])
            for typeattr in typeattrs:
                this_type.typeattrs.append(typeattr)
            child_type = this_type

        if this_type.parent_id is not None and get_parent_types is True:
            return self.parent.get_type(this_type.parent_id,
                                         child_type=child_type,
                                         get_parent_types=get_parent_types)
        child_type.ta_tree = None
        return child_type


    @staticmethod
    def get_type_parent_ids(ttype_id):
        """
            Returns a list of the ids of TemplateTypes which
            are parents of the TemplateType specified in the
            <ttype_id> argument.
        """
        q = get_session().query(TemplateType)
        q = q.filter(TemplateType.id == ttype_id)
        q = q.options(noload(TemplateType.typeattrs))
        try:
            ttype = q.one()
        except sqlalchemy.exc.NoResultFound:
            raise HydraError(f"No TemplateType found with id {ttype_id}")

        parent_ids = []

        while ttype.parent_id:
            parent_ids.insert(0, ttype.parent_id)
            ttype = get_session().query(TemplateType).filter(TemplateType.id == ttype.parent_id).one()

        return parent_ids


    def get_types(self, type_tree={}, child_types=None, get_parent_types=True, child_template_id=None):
        """
            Return all the templatetypes relevant to this template.
            If this template inherits from another, look up the tree to compile
            an exhaustive list of types, removing any duplicates, prioritising
            the ones closest to this template (my immediate parent's values are used instead
            of its parents)

        """
        log.info("Getting Template Types..")

        #This avoids python's mutable keyword arguments causing child_data to keep its values between
        #function calls
        if child_types is None:
            child_types = []
            type_tree = {}

        #Add resource attributes which are not defined already
        types_i = get_session().query(TemplateType).filter(
            TemplateType.template_id == self.id).options(noload(TemplateType.typeattrs)).all()
        types = [JSONObject(t) for t in types_i]

        if child_template_id is None:
            child_template_id = self.id

        for i, this_type in enumerate(types):
            if this_type.parent_id is not None:
                this_type.parent_ids = Template.get_type_parent_ids(this_type.id)

            this_type.child_template_id = child_template_id

            #This keeps track of which type attributes are currently associated
            #to this type. We'll use the data in this dict to set the 'typeattrs'
            #at the end
            if not hasattr(this_type, 'ta_tree') or this_type.ta_tree is None:
                this_type.ta_tree = {}

            #get all the type attrs for this type, and add any which are missing
            typeattrs_i = get_session().query(TypeAttr)\
                .filter(TypeAttr.type_id == this_type.id)\
                .options(joinedload(TypeAttr.attr))\
                .options(joinedload(TypeAttr.default_dataset)).all()

            typeattrs = [JSONObject(ta) for ta in typeattrs_i]

            #Is this type the parent of a type. If so, we don't want to add a new type
            #we want to update an existing one with any data that it's missing
            if this_type.id in type_tree:
                #This is a deleted type, so ignore it in the parent
                if type_tree[this_type.id] is  None:
                    continue

                #Find the child type and update it.
                child_type = type_tree[this_type.id]

                child_type = self.set_inherited_columns(this_type, child_type, types_i[i])

                #check the child's typeattrs. If a typeattr exists on the parent, with the
                #same attr_id, then it should be ignore. THis can happen when adding a typeattr
                #to the child first, then the parent
                child_typeattrs = [ta.attr_id for ta in child_type.typeattrs]

                for typeattr in typeattrs:
                    if typeattr.attr_id in child_typeattrs:
                        log.debug("Found a typeattr for attribute %s on the "
                             "child type %s (%s). Ignoring",
                             typeattr.attr_id, child_type.name, child_type.id)
                        continue

                    #Does this typeattr have a child?
                    child_typeattr = type_tree[this_type.id].ta_tree.get(typeattr.id)
                    if child_typeattr is None:

                        #there is no child, so check if it is a child
                        if typeattr.parent_id is not None:
                            #it has a parent, so add it to the type's tree dict
                            #for processing farther up the tree
                            type_tree[this_type.id].ta_tree[typeattr.parent_id] = typeattr
                        child_type.typeattrs.append(typeattr)
                    else:
                        child_typeattr = self.set_inherited_columns(typeattr, child_typeattr, types_i[i])


                if this_type.parent_id is not None:
                    type_tree[this_type.parent_id] = child_type

            else:
                if not hasattr(this_type, 'typeattrs'):
                    setattr(this_type, 'typeattrs', [])
                for typeattr in typeattrs:
                    #is this a child? if so, register it as one
                    if typeattr.parent_id is not None:
                        this_type.ta_tree[typeattr.parent_id] = typeattr
                    this_type.typeattrs.append(typeattr)

                child_types.append(this_type)
                #set
                if this_type.parent_id is not None:
                    type_tree[this_type.parent_id] = this_type


        if self.parent is not None and get_parent_types is True:
            return self.parent.get_types(type_tree=type_tree,
                                         child_types=child_types,
                                         get_parent_types=get_parent_types,
                                        child_template_id=child_template_id)

        #clean up
        for child_type in child_types:
            child_type.ta_tree = None

        return child_types

    def get_hierarchy(self, user_id):

        hierarchy = [JSONObject(self)]
        if self.parent_id:
            hierarchy = hierarchy + self.parent.get_hierarchy(user_id)
        return hierarchy

class TemplateType(Base, Inspect):
    """
    Template Type
    """

    __tablename__ = 'tTemplateType'
    __table_args__ = (
        UniqueConstraint('template_id', 'name', 'resource_type', name="unique type name"),
    )

    #these are columns which are not allowed to be changed by a child type
    _protected_columns = ['id', 'template_id', 'parent_id', 'cr_date', 'updated_at']
    _hierarchy_columns = ['name', 'resource_type']

    id = Column(Integer(), primary_key=True, nullable=False)
    parent_id = Column(Integer(), ForeignKey('tTemplateType.id'))
    template_id = Column(Integer(), ForeignKey('tTemplate.id'), nullable=False)
    name = Column(String(200), nullable=True)
    description = Column(String(1000))
    resource_type = Column(String(200), nullable=True)
    alias = Column(String(100))
    status = Column(String(1),  nullable=True)
    layout = Column(Text().with_variant(mysql.LONGTEXT, 'mysql'))
    cr_date = Column(TIMESTAMP(), nullable=False, server_default=text(u'CURRENT_TIMESTAMP'))

    parent = relationship('TemplateType', remote_side=[id],
                          backref=backref("children", order_by=id))

    template = relationship('Template',
                            backref=backref("templatetypes",
                                            order_by=id,
                                            cascade="all, delete-orphan"))

    _parents = ['tTemplate']
    _children = ['tTypeAttr']

    def get_typeattrs(self, ta_tree={}, child_typeattrs=None, get_parent_types=True):
        """
            This is unfinished
        """
        #This avoids python's mutable keyword arguments causing child_data to keep its values between
        #function calls
        if child_typeattrs is None:
            child_typeattrs = []
            ta_tree = {}

        #get all the type attrs for this type, and add any which are missing
        typeattrs_i = get_session().query(TypeAttr)\
            .filter(TypeAttr.type_id == self.id)\
            .options(joinedload(TypeAttr.default_dataset)).all()
        typeattrs = [JSONObject(ta) for ta in typeattrs_i]


        for i, typeattr in enumerate(typeattrs):

            #Does this typeattr have a child?
            child_typeattr = ta_tree.get(typeattr.id)
            if child_typeattr is None:
                #there is no child, so check if it is a child
                if typeattr.parent_id is not None:
                    #it has a parent, so add it to the tree dict
                    #for processing farther up the tree
                    ta_tree[typeattr.parent_id] = typeattr

            else:
                child_typeattr = self.template.set_inherited_columns(typeattr, child_typeattr, typeattrs_i[i])
                child_typeattrs.append(typeattr)



        if self.parent is not None and get_parent_types is True:
            return self.parent.get_typeattrs(ta_tree=ta_tree,
                                              child_typeattrs=child_typeattrs,
                                              get_parent_types=get_parent_types)
        return child_typeattrs

    def get_children(self):
        """
            Get the child types of a template type
        """

        child_types = get_session().query(TemplateType)\
                .filter(TemplateType.parent_id==self.id).all()
        return child_types

    def check_can_delete_resourcetypes(self, delete_resourcetypes=False):
        """
            Check if the delete operation will allow the deletion
            of resourcetypes. Default is NO
        """
        #Check if there are any resourcetypes associated to this type. If so,
        #don't delete it.
        resourcetype_count = get_session().query(ResourceType.id)\
                .filter(ResourceType.type_id == self.id).count()

        if resourcetype_count > 0 and delete_resourcetypes is False:
            raise HydraError(f"Unable to delete type. Template Type {self.id} has "
                             f"{resourcetype_count} resources associated to it. "
                             "Use the 'delete_resourcetypes' flag to delete these also.")

    def delete_children(self, delete_resourcetypes=False):
        """
            Delete the children associated to this type.
            THIS SHOULD BE DONE WITH EXTREME CAUTION.
            args:
                delete_resourcetypes (bool): If any resourcetypes are found to be
                associated to a child, throw an error to avoid leaving nodes with no types.
                If this flag is is set to true, then delete the resourcetypes

            This function works its way all the way down the tree to the leaf nodes
            and then deletes them from the leaf to the source
        """

        children = self.get_children()

        for child in children:
            child.delete_children(delete_resourcetypes=delete_resourcetypes)

            child.check_can_delete_resourcetypes(delete_resourcetypes=delete_resourcetypes)
            #delete all the resource types associated to this type
            if delete_resourcetypes is True:
                self.delete_resourcetypes()

            #delete the typeattrs
            for ta in child.typeattrs:
                get_session().delete(ta)

            get_session().delete(child)

    def delete_resourcetypes(self):
        """
        Delete the resourcetypes associated to a type
        """
        type_rs = get_session().query(ResourceType).filter(ResourceType.type_id==self.id).all()

        log.warning("Forcing the deletion of %s resource types from type %s",\
                 len(type_rs), self.id)

        for resource_type in type_rs:
            get_session().delete(resource_type)

class TypeAttr(Base, Inspect):
    """
        Type Attribute
    """

    __tablename__ = 'tTypeAttr'

    __table_args__ = (
        UniqueConstraint('type_id', 'attr_id', name='type_attr_1'),
    )

    id = Column(Integer(), primary_key=True, nullable=False)
    parent_id = Column(Integer(), ForeignKey('tTypeAttr.id'))
    attr_id = Column(Integer(), ForeignKey('tAttr.id'), nullable=False)
    type_id = Column(Integer(), ForeignKey('tTemplateType.id', ondelete='CASCADE'),
                     nullable=False)
    default_dataset_id = Column(Integer(), ForeignKey('tDataset.id'))
    attr_is_var = Column(String(1), server_default=text(u"'N'"))
    data_type = Column(String(60))
    data_restriction = Column(Text().with_variant(mysql.LONGTEXT, 'mysql'))
    unit_id = Column(Integer(), ForeignKey('tUnit.id'))
    description = Column(String(1000))
    properties = Column(Text().with_variant(mysql.LONGTEXT, 'mysql'))
    status = Column(String(1),  nullable=True)
    cr_date = Column(TIMESTAMP(), nullable=False, server_default=text(u'CURRENT_TIMESTAMP'))


    parent = relationship('TypeAttr', remote_side=[id], backref=backref("children", order_by=id))

    attr = relationship('Attr')
    #Don't use a cascade delete all here. Instead force the code to delete the typeattrs
    #manually, to avoid accidentally deleting them
    templatetype = relationship('TemplateType',
                                backref=backref("typeattrs",
                                                order_by=attr_id))
    unit = relationship('Unit',
                        backref=backref("typeattr_unit",
                                        order_by=unit_id))
    default_dataset = relationship('Dataset')

    _parents = ['tTemplateType', 'tUnit']
    _children = []

    def get_attr(self):
        """
            Get the attribute object
        """
        attr = None
        try:
            self.attr

            if self.attr is not None:
                attr = self.attr
        except:
            log.info("Unable to lazy-load attribute on typeattr %s", self.id)

        if attr is None:
            attr = get_session().query(Attr).filter(Attr.id == self.attr_id).first()

        return attr

    def get_unit(self):
        """
            Get the unit object
        """
        unit = None
        try:
            self.unit

            if self.unit is not None:
                unit = self.unit
        except:
            log.info("Unable to lazy-load unitibute on typeunit %s", self.id)

        if unit is None:
            unit = get_session().query(Unit).filter(unit.id == self.unit_id).first()

        return unit



    @property
    def is_var(self):
        return self.attr_is_var

    def get_properties(self):
        """
            This is here to match the JSONObject TypeAttr class which also
            has a get_properties, and which is required by some functions where
            both types can be validly passed in (like _set_typeattr)
        """
        return self.properties

class ResourceType(Base, Inspect):
    """
    """

    __tablename__='tResourceType'
    __table_args__ = (
        UniqueConstraint('network_id', 'type_id', name='net_type_1'),
        UniqueConstraint('node_id', 'type_id', name='node_type_1'),
        UniqueConstraint('link_id', 'type_id',  name = 'link_type_1'),
        UniqueConstraint('group_id', 'type_id', name = 'group_type_1'),

    )
    id = Column(Integer, primary_key=True, nullable=False)
    type_id = Column(Integer(), ForeignKey('tTemplateType.id'), primary_key=False, nullable=False)
    #This template id is used when the template and the type are not from the same template
    #i.e. the resource type is being used in a child template
    #If null, then the resources has either been created using a non-child template, or with a resource
    #type in a child template which has been entered to the DB, because the parent type has been altered in the child
    child_template_id = Column(Integer(), ForeignKey('tTemplate.id'), primary_key=False, nullable=True)
    ref_key = Column(String(60),nullable=False)
    network_id  = Column(Integer(),  ForeignKey('tNetwork.id'), nullable=True,)
    node_id     = Column(Integer(),  ForeignKey('tNode.id'), nullable=True)
    link_id     = Column(Integer(),  ForeignKey('tLink.id'), nullable=True)
    group_id    = Column(Integer(),  ForeignKey('tResourceGroup.id'), nullable=True)
    cr_date = Column(TIMESTAMP(),  nullable=False, server_default=text(u'CURRENT_TIMESTAMP'))

    #Don't used a delete cascade here because deleting the type accidentally can delete
    #this data. INstead the resource types should be deleted manually before the deletion
    #of the type
    templatetype = relationship('TemplateType', backref=backref('resourcetypes', uselist=True))

    network = relationship('Network', backref=backref('types', uselist=True, cascade="all, delete-orphan"), uselist=False)
    node = relationship('Node', backref=backref('types', uselist=True, cascade="all, delete-orphan"), uselist=False)
    link = relationship('Link', backref=backref('types', uselist=True, cascade="all, delete-orphan"), uselist=False)
    resourcegroup = relationship('ResourceGroup', backref=backref('types', uselist=True, cascade="all, delete-orphan"), uselist=False)

    _parents  = ['tNode', 'tLink', 'tResourceGroup', 'tNetwork', 'tProject']
    _children = []

    def get_resource(self):
        ref_key = self.ref_key
        if ref_key == 'PROJECT':
            return self.project
        elif ref_key == 'NETWORK':
            return self.network
        elif ref_key == 'NODE':
            return self.node
        elif ref_key == 'LINK':
            return self.link
        elif ref_key == 'GROUP':
            return self.group

    def get_resource_id(self):
        ref_key = self.ref_key
        if ref_key == 'PROJECT':
            return self.project_id
        elif ref_key == 'NETWORK':
            return self.network_id
        elif ref_key == 'NODE':
            return self.node_id
        elif ref_key == 'LINK':
            return self.link_id
        elif ref_key == 'GROUP':
            return self.group_id

    def get_templatetype(self):
        """
            If this type is that of a child templatetype, then the full type
            needs to be constructed by the template. So instead of getting the
            template type directly, we get the template, then request the type.
        """


        #This resource was created using a child template
        if self.child_template_id is not None:
            template_i = get_session().query(Template)\
                .filter(Template.id == self.child_template_id).one()
        else:
            type_i = get_session().query(TemplateType).filter(TemplateType.id == self.type_id).one()

            if type_i.parent_id is None:
                return type_i

            template_i = get_session().query(Template)\
                    .filter(Template.id == type_i.template_id).one()

        type_i = template_i.get_type(self.type_id)

        return JSONObject(type_i)



import pytest
import json
import datetime
import collections

import hydra_base as hb
import pandas as pd

from hydra_base.exceptions import HydraError
from hydra_base.lib.HydraTypes.Registry import HydraObjectFactory
from hydra_base.lib.HydraTypes.Types import Scalar, Array

import logging
log = logging.getLogger("objects")


def generator(size):
    """
        To avoid python 2/3 compatibility issues, use a custom generator instead of 'range' directly
    """
    for i in range(size):
        yield i

""" Type arguments """

scalar_valid_values       = [ 46, -1, 0, 7.7, -0.0 ]
scalar_invalid_values     = [ "one", pd, {} ]

array_valid_values        = [ [-2, -1, 0, 1, 2], list(range(32)), [ 0.5e-3, 0.5, 0.5e3 ] ]
array_invalid_values      = [ generator(32), 77, {}, "justastring" ]

timeseries_valid_values   = [ {"0": {"1979 Feb 2 0100":7, "01:00 2 Feb 1979":9}}, {"0":{"2012":12, "2013":13, "2014":14}}, {"0":{"18:00 31 August 1977":100}} ]
timeseries_invalid_values = [ "otheriterable", list(range(12)), {"JAN":1, "FEB":2, "MAR":3, "APR":4}, set(), ["01:00 30 Feb 1979"] ]

dataframe_valid_values    = [ {"data" : {"fr": "ame"}}, {"one": ["first"], "two": ["second"]}, {"n":{"e":{"s":{"t":{"e":"d"}}}}} ]
dataframe_invalid_values  = [ 77, set(), {"one": "first", "two": "second"} ]


""" Scalar type tests """

@pytest.mark.parametrize("value", scalar_valid_values)
def test_create_scalar(value):
    scalar_dataset = hb.lib.objects.Dataset({'type':'scalar', 'value': value})
    value = scalar_dataset.parse_value()


@pytest.mark.parametrize("value", scalar_invalid_values)
def test_fail_create_scalar(value):
    with pytest.raises(HydraError):
        scalar_dataset = hb.lib.objects.Dataset({'type':'scalar', 'value': value})
        value = scalar_dataset.parse_value()


@pytest.fixture(params=scalar_valid_values)
def make_scalar(request):
    scalar_dataset = hb.lib.objects.Dataset({'type':'scalar', 'value': request.param})
    o = HydraObjectFactory.fromDataset(scalar_dataset.type, scalar_dataset.value)
    Fixret = collections.namedtuple("fixret", ("obj", "fixarg"))
    yield Fixret(o, request.param)     # Expose current fixture arg to test


def test_raw_scalar(make_scalar):
    assert type(make_scalar.obj) == Scalar
    assert make_scalar.obj.tag == Scalar.tag
    assert make_scalar.obj.value is not None
    assert make_scalar.obj.value == str(make_scalar.fixarg)
    assert make_scalar.obj.validate() is None



""" Array type tests """

@pytest.mark.parametrize("value", array_valid_values)
def test_create_array(value):
    array_dataset = hb.lib.objects.Dataset({'type':'array', 'value': json.dumps(value)})
    value = array_dataset.parse_value()


@pytest.mark.parametrize("value", array_invalid_values)
def test_fail_create_array(value):
    with pytest.raises( (HydraError, TypeError) ):
        print(value)
        array_dataset = hb.lib.objects.Dataset({'type':'array', 'value': json.dumps(value)})
        value = array_dataset.parse_value()


@pytest.fixture(params=array_valid_values)
def make_array(request):
    array_dataset = hb.lib.objects.Dataset({'type':'array', 'value': json.dumps(request.param) })
    o = HydraObjectFactory.fromDataset(array_dataset.type, array_dataset.value)
    Fixret = collections.namedtuple("fixret", ("obj", "fixarg"))
    yield Fixret(o, request.param)     # Expose current fixture arg to test


def test_raw_array(make_array):
    assert type(make_array.obj) == Array
    assert make_array.obj.tag == Array.tag
    assert make_array.obj.value is not None
    assert make_array.obj.value == str(make_array.fixarg)
    assert make_array.obj.validate() is None


""" Timeseries type tests """

@pytest.mark.parametrize("value", timeseries_valid_values)
def test_create_timeseries(value):
    timeseries_dataset = hb.lib.objects.Dataset({'type':'timeseries', 'value': json.dumps(value)})
    value = timeseries_dataset.parse_value()


@pytest.mark.parametrize("value", timeseries_invalid_values)
def test_fail_create_timeseries(value):
    with pytest.raises( (HydraError, TypeError, ValueError) ):
        timeseries_dataset = hb.lib.objects.Dataset({'type':'timeseries', 'value': json.dumps(value)})
        value = timeseries_dataset.parse_value()

""" DataFrame type tests """

@pytest.mark.parametrize("value", dataframe_valid_values)
def test_create_dataframe(value):
    dataframe_dataset = hb.lib.objects.Dataset({'type':'dataframe', 'value': json.dumps(value)})
    value = dataframe_dataset.parse_value()


@pytest.mark.parametrize("value", dataframe_invalid_values)
def test_fail_create_dataframe(value):
    with pytest.raises( (HydraError, TypeError, ValueError) ):
        dataframe_dataset = hb.lib.objects.Dataset({'type':'dataframe', 'value': json.dumps(value)})
        value = dataframe_dataset.parse_value()


def test_dataframe_order_preserved():
    #make the index deliberately non-ordered
    index = ['0', '2', '1']
    cols = ['A']
    data = ['x', 'y', 'z']

    df = pd.DataFrame(data, columns=cols, index=index)

    dataframe_dataset = hb.lib.objects.Dataset({'type':'dataframe', 'value': df.to_json()})
    parsed_value = dataframe_dataset.parse_value()

    assert parsed_value == df.to_json()


def test_dataframe_time_index_order_preserved():
    #Test to make sure date time indices remain in the order they are sent to
    #hydra, if they are not ordered by time for whatever reason.

    #make the index deliberately non-ordered
    index = ['2010-02-01', '2005-01-01', '2012-01-01']
    cols = ['A']
    data = ['x', 'y', 'z']

    df = pd.DataFrame(data, columns=cols, index=index)

    dataframe_dataset = hb.lib.objects.Dataset({'type':'dataframe', 'value': df.to_json()})
    parsed_value = dataframe_dataset.parse_value()

    assert parsed_value == df.to_json()
